#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict, List
from ads.common import auth as authutil
from ads.data_labeling.reader.dataset_reader import LabeledDatasetReader
from ads.data_labeling.visualizer import image_visualizer, text_visualizer
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)

ROWS_TO_RENDER_LIMIT = 50


class DataLabelingAccessMixin:
    """Mixin class for labeled text data."""

    @staticmethod
    def read_labeled_data(
        path: str = None,
        dataset_id: str = None,
        compartment_id: str = None,
        auth: Dict = None,
        materialize: bool = False,
        encoding: str = "utf-8",
        include_unlabeled: bool = False,
        format: str = None,
        chunksize: int = None,
    ):
        """Loads the dataset generated by data labeling service from either the export file or the Data Labeling Service.

        Parameters
        ----------
        path: (str, optional). Defaults to None
            The export file path, can be either local or object storage path.
        dataset_id: (str, optional). Defaults to None
            The dataset OCID.
        compartment_id: str. Defaults to the compartment_id from the env variable.
            The compartment OCID of the dataset.
        auth: (dict, optional). Defaults to None
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        materialize: (bool, optional). Defaults to False
            Whether the content of the dataset file should be loaded or it should return the file path to the content.
            By default the content will not be loaded.
        encoding: (str, optional). Defaults to 'utf-8'
            Encoding of files. Only used for "TEXT" dataset.
        include_unlabeled: (bool, optional). Default to False
            Whether to load the unlabeled records or not.
        format: (str, optional). Defaults to None
            Output format of annotations. Can be None, "spacy" for dataset
            Entity Extraction type or "yolo for Object Detection type.

            - When None, it outputs List[NERItem] or List[BoundingBoxItem],
            - When "spacy", it outputs List[Tuple],
            - When "yolo", it outputs List[List[Tuple]].

        chunksize: (int, optional). Defaults to None
            The amount of records that should be read in one iteration.
            The result will be returned in a generator format.

        Returns
        -------
        Union[Generator[pd.DataFrame, Any, Any], pd.DataFrame]
            `pd.Dataframe` if `chunksize` is not specified.
            `Generator[pd.Dataframe]` if `chunksize` is specified.

        Examples
        --------
        >>> import pandas as pd
        >>> import ads
        >>> from ads.common import auth as authutil
        >>> df = pd.DataFrame.ads.read_labeled_data(path="path_to_your_metadata.jsonl",
        ...                                         auth=authutil.api_keys(),
        ...                                         materialize=False)
                                    Path       Content               Annotations
            --------------------------------------------------------------------
            0   path/to/the/content/file                                     yes
            1   path/to/the/content/file                                      no

        >>> df = pd.DataFrame.ads.read_labeled_data_from_dls(dataset_id="your_dataset_ocid",
        ...                                                  compartment_id="your_compartment_id",
        ...                                                  auth=authutil.api_keys(),
        ...                                                  materialize=False)
                                    Path       Content               Annotations
            --------------------------------------------------------------------
            0   path/to/the/content/file                                     yes
            1   path/to/the/content/file                                      no
        """
        auth = auth or authutil.default_signer()

        if not path and not dataset_id:
            raise ValueError("Either `path` or the `dataset_id` need to be provided.")
        if path and dataset_id:
            raise ValueError("Only one of `path` and `dataset_id` need to be provided.")
        if path:
            reader = LabeledDatasetReader.from_export(
                path=path,
                auth=auth,
                encoding=encoding,
                materialize=materialize,
                include_unlabeled=include_unlabeled,
            )
        else:
            reader = LabeledDatasetReader.from_DLS(
                dataset_id=dataset_id,
                compartment_id=compartment_id,
                auth=auth,
                encoding=encoding,
                materialize=materialize,
                include_unlabeled=include_unlabeled,
            )

        ds_info = reader.info()
        if chunksize:

            def _bulk_read():
                for df in reader.read(format=format, chunksize=chunksize):
                    df.ads._info = ds_info
                    yield df

            return _bulk_read()

        df = reader.read(format=format)
        df.ads._info = ds_info
        return df

    def render_bounding_box(
        self,
        options: Dict = None,
        content_column: str = "Content",
        annotations_column: str = "Annotations",
        categories: List[str] = None,
        limit: int = ROWS_TO_RENDER_LIMIT,
        path: str = None,
    ) -> None:
        """Renders bounding box dataset. Displays only first 50 rows.

        Parameters
        ----------
        options: dict
            The colors options specified for rendering.
        content_column: Optional[str]
            The column name with the content data.
        annotations_column: Optional[str]
            The column name for the annotations list.
        categories: Optional List[str]
            The list of object categories in proper order for model training.
            Only used when bounding box annotations are in YOLO format.
            Example: ['cat','dog','horse']
        limit: Optional[int]. Defaults to 50
            The maximum amount of records to display.
        path: Optional[str]
            Path to save the image with annotations to local directory.

        Returns
        -------
        None
            Nothing

        Examples
        --------
        >>> import pandas as pd
        >>> import ads
        >>> from ads.common import auth as authutil
        >>> df = pd.DataFrame.ads.read_labeled_data(path="path_to_your_metadata.jsonl",
        ...                                         auth=authutil.api_keys(),
        ...                                         materialize=True)
        >>> df.ads.render_bounding_box(content_column="Content", annotations_column="Annotations")
        """
        items = image_visualizer._df_to_bbox_items(
            self._obj.iloc[0:limit] if self._obj.shape[0] > limit else self._obj,
            content_column=content_column,
            annotations_column=annotations_column,
            categories=categories,
        )
        image_visualizer.render(items, options=options, path=path)

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def render_ner(
        self,
        options: Dict = None,
        content_column: str = "Content",
        annotations_column: str = "Annotations",
        limit: int = ROWS_TO_RENDER_LIMIT,
    ) -> None:
        """Renders NER dataset. Displays only first 50 rows.

        Parameters
        ----------
        options: dict
            The colors options specified for rendering.
        content_column: Optional[str]
            The column name with the content data.
        annotations_column: Optional[str]
            The column name for the annotations list.
        limit: Optional[int]. Defaults to 50
            The maximum amount of records to display.

        Returns
        -------
        None
            Nothing

        Examples
        --------
        >>> import pandas as pd
        >>> import ads
        >>> from ads.common import auth as authutil
        >>> df = pd.DataFrame.ads.read_labeled_data(path="path_to_your_metadata.jsonl",
        ...                                         auth=authutil.api_keys(),
        ...                                         materialize=True)
        >>> df.ads.render_ner(content_column="Content", annotations_column="Annotations")
        """
        items = text_visualizer._df_to_ner_items(
            self._obj.iloc[0:limit] if self._obj.shape[0] > limit else self._obj,
            content_column=content_column,
            annotations_column=annotations_column,
        )
        result_html = text_visualizer.render(items=items, options=options)

        from IPython.core.display import HTML, Markdown, display

        display(Markdown(result_html))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/
import copy
import json
from typing import Union, List, Dict
from configparser import ConfigParser, ExtendedInterpolation
from configparser import (
    InterpolationSyntaxError,
    InterpolationDepthError,
)
from configparser import NoSectionError, NoOptionError, MAX_INTERPOLATION_DEPTH


class EnvVarInterpolation(ExtendedInterpolation):
    """Modified version of ExtendedInterpolation to ignore errors

    https://github.com/python/cpython/blob/main/Lib/configparser.py
    """

    def before_set(self, parser, section: str, option: str, value: str) -> str:
        return value

    def _interpolate_some(self, parser, option, accum, rest, section, map, depth):
        rawval = parser.get(section, option, raw=True, fallback=rest)
        if depth > MAX_INTERPOLATION_DEPTH:
            raise InterpolationDepthError(option, section, rawval)
        while rest:
            p = rest.find("$")
            if p < 0:
                accum.append(rest)
                return
            if p > 0:
                accum.append(rest[:p])
                rest = rest[p:]
            # p is no longer used
            c = rest[1:2]
            if c == "$":
                accum.append("$")
                rest = rest[2:]
            elif c == "{":
                m = self._KEYCRE.match(rest)
                if m is None:
                    raise InterpolationSyntaxError(
                        option,
                        section,
                        "Bad interpolation variable reference %r" % rest,
                    )
                path = m.group(1).split(":")
                sect = section
                opt = option
                try:
                    if len(path) == 1:
                        opt = parser.optionxform(path[0])
                        v = map[opt]
                    elif len(path) == 2:
                        sect = path[0]
                        opt = parser.optionxform(path[1])
                        v = parser.get(sect, opt, raw=True)
                    else:
                        raise InterpolationSyntaxError(
                            option,
                            section,
                            "More than one ':' found: %r" % (rest[m.end() :],),
                        )
                except (KeyError, NoSectionError, NoOptionError):
                    accum.append(rest)
                    return

                rest = rest[m.end() :]
                if "$" in v:
                    self._interpolate_some(
                        parser,
                        opt,
                        accum,
                        v,
                        sect,
                        dict(parser.items(sect, raw=True)),
                        depth + 1,
                    )
                else:
                    accum.append(v)
            else:
                accum.append("$")
                rest = rest[1:]


def parse(env_var: Union[Dict, List[dict]]) -> dict:
    """Parse the environment variables and perform substitutions.
    This will also converts kubernetes style environment variables from a list to a dictionary.

    Parameters
    ----------
    env_var : dict or list
        Environment variables specified as a list or a dictionary.
        If evn_var is a list, it should be in the format of:
            "[{"name": "ENV_NAME_1", "value": "ENV_VALUE_1"}, {"name": "ENV_NAME_2", "value": "ENV_VALUE_2"}]

    Returns
    -------
    dict
        Environment variable as a dictionary.
    """
    # Convert kubernetes style env to dict
    if isinstance(env_var, list):
        env_var = {ev["name"]: ev["value"] for ev in env_var}
    else:
        env_var = copy.deepcopy(env_var)
    config = ConfigParser(interpolation=EnvVarInterpolation())
    config.optionxform = str
    for k in env_var.keys():
        if env_var[k] is None:
            # Convert None to empty string
            env_var[k] = ""
        elif not isinstance(env_var[k], str):
            # If the value is not a string,
            # try to dump it as json string
            try:
                env_var[k] = json.dumps(env_var[k])
            except Exception:
                # Cast the value to string if it is not json serializable
                env_var[k] = str(env_var[k])

    config["envs"] = env_var
    return {k: config["envs"].get(k) for k in env_var.keys()}


def escape(s: str) -> str:
    return s.replace("$", "$$")



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import

import json
import re
import copy
from collections import defaultdict
from time import time

import numpy as np
import pandas as pd
from sklearn.utils import Bunch
from sklearn.feature_extraction.text import CountVectorizer

from ads.common import utils, logger
from ads.common.card_identifier import card_identify
from ads.common.utils import JsonConverter
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)


class TypedFeature(Bunch):
    def __init__(self, name, meta_data):
        assert "type" in meta_data
        self.meta_data = meta_data
        self.meta_data["feature_name"] = name
        super().__init__(**meta_data)

    @staticmethod
    def build(name, series):
        pass

    def __repr__(self):
        d = copy.deepcopy(self.meta_data)
        if "internal" in d:
            d.pop("internal")
        return json.dumps(d, indent=2, cls=JsonConverter)


class ContinuousTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    @runtime_dependency(module="scipy", install_from=OptionalDependency.VIZ)
    def build(name, series):
        series = series.astype("float")

        non_null_series = series.loc[~series.isnull()]
        desc = non_null_series.describe()
        stats = {
            "mode": series.mode().iloc[0],
            "median": desc["50%"],
            "kurtosis": series.kurt(),
            "variance": series.var(),
            "skewness": series.skew(),
            "outlier_percentage": 100
            * np.count_nonzero(np.abs(scipy.stats.zscore(non_null_series) >= 2.999))
            / series.size,
        }
        stats.update({k: v for k, v in desc.items()})

        return ContinuousTypedFeature(
            name,
            {
                "type": "continuous",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": stats,
                "internal": {"sample": series.head(5)},
            },
        )


class DiscreteTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)


class OrdinalTypedFeature(DiscreteTypedFeature):
    def __init__(self, name, meta_data):
        DiscreteTypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        nulls_removed = series.astype("category").loc[~series.isnull()]
        desc = nulls_removed.describe(include="all")
        value_counts = series.value_counts(ascending=False)
        x_min, x_max = np.nanmin(nulls_removed), np.nanmax(nulls_removed)

        stats = {
            "unique percentage": 100 * desc["unique"] / desc["count"],
            "x_min": x_min,
            "x_max": x_max,
            "mode": series.mode().iloc[0],
        }
        stats.update({k: v for k, v in desc.items()})

        return OrdinalTypedFeature(
            name,
            {
                "type": "ordinal",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": stats,
                "internal": {
                    "sample": series.head(5),
                    "unique": desc["unique"],
                    "counts": utils.truncate_series_top_n(
                        value_counts, n=min(16, len(value_counts))
                    ),
                    "high_cardinality": bool(desc["unique"] > 30),
                    "very_high_cardinality": bool(
                        desc["unique"] >= 0.95 * desc["count"]
                    ),
                },
            },
        )


class CategoricalTypedFeature(DiscreteTypedFeature):
    def __init__(self, name, meta_data):
        DiscreteTypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        desc = series.astype("category").loc[~series.isnull()].describe(include="all")
        value_counts = series.value_counts(ascending=False)
        if isinstance(desc["top"], str):
            mode = desc["top"] if len(desc["top"]) < 30 else desc["top"][:24] + "..."
        else:
            mode = desc["top"]

        stats = {
            "unique percentage": 100 * desc["unique"] / desc["count"],
            "mode": mode,
        }
        stats.update({k: v for k, v in desc.items()})

        return CategoricalTypedFeature(
            name,
            {
                "type": "categorical",
                "low_level_type": series.dtype.name,
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "stats": stats,
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "unique": desc["unique"],
                    "counts": utils.truncate_series_top_n(
                        value_counts, n=min(16, len(value_counts))
                    ),
                    "high_cardinality": bool(desc["unique"] > 30),
                    "very_high_cardinality": bool(
                        desc["unique"] >= 0.95 * desc["count"]
                    ),
                },
            },
        )


class IPAddressTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        desc = series.astype("category").loc[~series.isnull()].describe()

        nulls_removed = series.loc[~series.isnull()]
        value_counts = nulls_removed.value_counts()

        return IPAddressTypedFeature(
            name,
            {
                "type": "ipaddress",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {
                    "unique percentage": 100 * desc["unique"] / desc["count"],
                    "mode": series.mode().iloc[0],
                },
                "internal": {
                    "sample": nulls_removed.sample(n=min(100, nulls_removed.size)),
                    "counts": utils.truncate_series_top_n(
                        value_counts, n=min(16, len(value_counts))
                    ),
                    "unique": desc["unique"],
                },
            },
        )


class PhoneNumberTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        desc = series.astype("category").loc[~series.isnull()].describe()

        nulls_removed = list(series.loc[~series.isnull()])
        pat = re.compile(r"^(\+?\d{1,2}[\s-])?\(?(\d{3})\)?[\s.-]?\d{3}[\s.-]?\d{4}$")
        area_codes = [
            g.group(2)
            for g in filter(None, [pat.match(x) for x in nulls_removed])
            if g and len(g.groups()) >= 2
        ]
        value_counts = pd.Series(area_codes).value_counts()

        # resolve area codes
        return PhoneNumberTypedFeature(
            name,
            {
                "type": "phone",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {
                    "unique percentage": 100 * desc["unique"] / desc["count"],
                    "mode": series.mode().iloc[0],
                },
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "counts": utils.truncate_series_top_n(
                        value_counts, n=min(16, len(value_counts))
                    ),
                    "unique": desc["unique"],
                },
            },
        )


class GISTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series, samples):
        desc = series.astype("category").loc[~series.isnull()].describe()
        value_counts = series.value_counts(ascending=False)
        return GISTypedFeature(
            name,
            {
                "type": "gis",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {
                    "observations": desc["count"],
                    "unique percentage": 100 * desc["unique"] / desc["count"]
                    # TODO mid point
                },
                "internal": {"sample": samples, "unique": desc["unique"]},
            },
        )


class AddressTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        nulls_removed = series.loc[~series.isnull()]
        mean_document_length = nulls_removed.str.len().mean()
        sampled_nulls_removed = nulls_removed.sample(n=min(len(nulls_removed), 1000))

        return AddressTypedFeature(
            name,
            {
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "type": "address",
                "low_level_type": series.dtype.name,
                "stats": {
                    "mean_character_count": mean_document_length,
                    "mean_word_length": nulls_removed.str.split().str.len().mean(),
                },
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "word_frequencies": DocumentTypedFeature.vectorization(
                        name, sampled_nulls_removed, mean_document_length
                    ),
                },
            },
        )


class DocumentTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    @runtime_dependency(module="wordcloud", install_from=OptionalDependency.TEXT)
    def corpus_processor(series):
        pat_punct = re.compile(r"[^a-zA-Z]", re.UNICODE + re.MULTILINE)
        pat_tags = re.compile(r"<.*?>", re.UNICODE + re.MULTILINE)
        pat_digits = re.compile(r"(\d|\W)+", re.UNICODE + re.MULTILINE)
        pat_splitter = re.compile(r"\s+", re.UNICODE + re.MULTILINE)
        for doc in series:
            # Remove punctuations, tags,  special characters and digits
            text = pat_punct.sub(" ", doc.lower())

            # remove tags
            text = pat_tags.sub("", text)

            # remove special characters and digits
            text = pat_digits.sub(" ", text)

            from wordcloud import STOPWORDS

            stop_words = STOPWORDS

            # Convert to list from string and return a generator expression
            yield " ".join(
                [
                    x
                    for x in pat_splitter.split(text)
                    if len(x) >= 3 and x not in stop_words
                ]
            )

    @staticmethod
    def sub_vectorization(
        feature_name,
        series,
        min_df=0.0,
        max_df=1.0,
        min_tf=2,
    ):
        start = time()

        v1 = CountVectorizer(
            ngram_range=(1, 1),
            max_features=2000,
            decode_error="ignore",
            strip_accents="unicode",
            stop_words=None,
            min_df=min_df,
            max_df=max_df,
        )
        X1 = v1.fit_transform(DocumentTypedFeature.corpus_processor(series))
        unigrams = {
            k: int(v)
            for k, v in dict(
                zip(v1.get_feature_names(), np.asarray(X1.sum(axis=0)).ravel())
            ).items()
        }

        v2 = CountVectorizer(
            ngram_range=(2, 2),
            max_features=2000,
            decode_error="ignore",
            strip_accents="unicode",
            stop_words=None,
            min_df=min_df,
            max_df=max_df,
        )
        X2 = v2.fit_transform(DocumentTypedFeature.corpus_processor(series))
        bigrams = {
            k: int(v)
            for k, v in dict(
                zip(v2.get_feature_names(), np.asarray(X2.sum(axis=0)).ravel())
            ).items()
        }

        # drop unigrams that are prefixes of bigrams
        for k, v in bigrams.items():
            # take the key (k), split it to get first word and look for that word in unigrams
            for word in k.split():
                if word in unigrams:
                    # boost the bigram
                    bigrams[k] = v + unigrams[word]
                    del unigrams[word]

        unigrams.update(bigrams)

        ret = {k: v for k, v in unigrams.items() if v >= min_tf}
        return ret if ret else unigrams.items()

    @staticmethod
    def vectorization(feature_name, series, mean_document_length):
        # set the min_df and max_df as functions of the mean_document_length
        min_df = 0.1
        max_df = 0.8

        if mean_document_length > 5000:
            min_df = 0.2
            max_df = 0.7
        try:
            return DocumentTypedFeature.sub_vectorization(
                feature_name, series, min_df=min_df, max_df=max_df
            )
        except ValueError:
            return DocumentTypedFeature.sub_vectorization(
                feature_name, series, min_df=0.0, max_df=1.0
            )

    @staticmethod
    def build(name, series, is_cjk, is_html):

        internal = {"cjk": is_cjk, "html": is_html}

        if is_cjk:
            stats = {"mean_document_character_count": int(series.str.len().mean())}

            internal["sample"] = series.sample(n=min(100, series.size))

        else:
            nulls_removed = series.loc[~series.isnull()]
            if is_html:
                nulls_removed = nulls_removed.replace(
                    to_replace="<[^<]+?>", value=" ", regex=True
                ).replace("\n", " ", regex=True)

            internal["sample"] = nulls_removed.sample(n=min(100, nulls_removed.size))

            mean_document_length = nulls_removed.str.len().mean()
            sample_size = int(
                2e6 // mean_document_length
            )  # vectorize at most 2MB of text

            sampled_nulls_removed = nulls_removed.sample(
                n=min(len(nulls_removed), sample_size)
            )

            stats = {
                "mean_document_character_count": int(series.str.len().mean()),
                "mean_word_length": nulls_removed.str.split().str.len().mean(),
            }

            internal["word_frequencies"] = DocumentTypedFeature.vectorization(
                name, sampled_nulls_removed, mean_document_length
            )

        return DocumentTypedFeature(
            name,
            {
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "type": "document",
                "low_level_type": series.dtype.name,
                "stats": stats,
                "internal": internal,
            },
        )


class ZipcodeTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        max_threshold = 24
        suffix = ""

        series = series.loc[~series.isnull()].astype(str).str.slice(0, 5)

        for i in [5, 4, 3, 2, 1]:
            if series.nunique() <= max_threshold:
                break
            series = series.str[:-1]
            suffix += "*"

        series += suffix

        return ZipcodeTypedFeature(
            name,
            {
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "type": "zipcode",
                "low_level_type": series.dtype.name,
                "stats": {"mode": series.mode().iloc[0]},
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "histogram": series.value_counts(),
                },
            },
        )


class UnknownTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        return UnknownTypedFeature(
            name,
            {
                "type": "unknown",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {"mode": series.mode().iloc[0]},
                "internal": {"sample": series.sample(n=min(100, series.size))},
            },
        )


class ConstantTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        desc = series.astype("category").loc[~series.isnull()].describe()
        if "top" in desc:
            value = desc["top"]
            value_counts = utils.truncate_series_top_n(series.value_counts())
        else:
            value = "NaN"
            value_counts = {value: len(series)}
        return ConstantTypedFeature(
            name,
            {
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "type": "constant",
                "stats": {"value": value},
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "counts": value_counts,
                },
            },
        )


class CreditCardTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        null_removed_series = series.loc[~series.isnull()]
        desc = null_removed_series.astype("category").describe()
        sampled_series = null_removed_series.sample(
            n=min(1000, len(null_removed_series))
        )

        d_scheme = defaultdict(int)

        # use counting method to build dict of credit card data
        for s in sampled_series:
            scheme = card_identify().identify_issue_network(s)
            d_scheme[scheme] += 1

        return CreditCardTypedFeature(
            name,
            {
                "type": "creditcard",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {
                    "unique percentage": 100 * desc["unique"] / desc["count"],
                    "mode": desc["top"],
                },
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "counts": dict(d_scheme),
                    "unique": desc["unique"],
                },
            },
        )


class DateTimeTypedFeature(TypedFeature):
    def __init__(self, name, meta_data):
        TypedFeature.__init__(self, name, meta_data)

    @staticmethod
    def build(name, series):
        desc = series.loc[~series.isnull()].describe()

        return DateTimeTypedFeature(
            name,
            {
                "type": "datetime",
                "missing_percentage": 100 * series.isna().sum() / series.size,
                "low_level_type": series.dtype.name,
                "stats": {
                    "unique percentage": 100 * desc["unique"] / desc["count"],
                    "first": desc["first"],
                    "last": desc["last"],
                    "mode": desc["top"],
                },
                "internal": {
                    "sample": series.sample(n=min(100, series.size)),
                    "unique": desc["unique"],
                },
            },
        )



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os

import matplotlib as mpl
import matplotlib.cm as cm
import numpy as np
import pandas as pd
from ads.common import utils as au
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.dataset.correlation import _validate_correlation_methods
from ads.dataset.helper import deprecate_default_value


class BokehHeatMap(object):
    """
    Generate a HeatMap or horizontal bar plot to compare features.
    """

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def __init__(self, ds):

        from bokeh.io import output_notebook
        from bokeh.palettes import BuPu

        output_notebook()

        self.ds = ds
        self.colormap = cm.get_cmap("BuPu")
        self.bokehpalette = [
            mpl.colors.rgb2hex(m) for m in self.colormap(np.arange(self.colormap.N))
        ]

    def debug(self):
        """
        Return True if in debug mode, otherwise False.
        """
        return au.is_debug_mode()

    def flatten_corr_matrix(self, corr_matrix):
        """
        Flatten a correlation matrix into a pandas Dataframe.

        Parameters
        ----------
        corr_matrix: Pandas Dataframe
          The correlation matrix to be flattened.

        Returns
        -------
        corr_flatten: Pandas DataFrame
            The flattened correlation matrix.

        """
        rows = corr_matrix.index.values.tolist()
        columns = corr_matrix.columns

        corr_flatten = pd.DataFrame(
            [(r, c, corr_matrix[r][c]) for c in columns for r in rows],
            columns=["x", "y", "corr"],
        )
        return corr_flatten

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def plot_heat_map(
        self,
        matrix,
        xrange: list,
        yrange: list,
        low: float = 1,
        high=1,
        title: str = None,
        tool_tips: list = None,
    ):
        """
        Plots a matrix as a heatmap.


        Parameters
        ----------
        matrix: Pandas Dataframe
          The dataframe to be plotted.
        xrange: List of floats
          The range of x values to plot.
        yrange: List of floats
          The range of y values to plot.
        low: float, Defaults to 1
          The color mapping value for "low" points.
        high: float, Defaults to 1
          The color mapping value for "high" points.
        title: str, Defaults to None
          The optional title of the heat map.
        tool_tips: list of str, Defaults to None
          An optional list of tool tips to include with the plot.

        Returns
        -------
        fig: matplotlib Figure
            A matplotlib heatmap figure object.
        """
        if self.debug():
            print(matrix)
        mapper = bokeh.models.LinearColorMapper(
            palette=self.bokehpalette, low=low, high=high
        )
        source = bokeh.models.ColumnDataSource(matrix)

        from bokeh.plotting import figure

        p = figure(
            title=title,
            x_range=xrange,
            y_range=yrange,
            toolbar_location="below",
            toolbar_sticky=False,
            plot_width=600,
            plot_height=600,
        )

        p.rect(
            x="x",
            y="y",
            width=1,
            height=1,
            source=source,
            fill_color={"field": "corr", "transform": mapper},
            line_color=None,
        )
        p.xaxis.major_label_orientation = "vertical"

        if tool_tips:
            p.add_tools(bokeh.models.HoverTool(tooltips=tool_tips))

        color_bar = bokeh.models.ColorBar(
            color_mapper=mapper,
            major_label_text_font_size="5pt",
            ticker=bokeh.models.BasicTicker(desired_num_ticks=8),
            formatter=bokeh.models.PrintfTickFormatter(format="%0.2f"),
            label_standoff=6,
            border_line_color=None,
            location=(0, 0),
        )
        p.add_layout(color_bar, "right")
        return p

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def plot_hbar(
        self,
        matrix,
        low: float = 1,
        high=1,
        title: str = None,
        tool_tips: list = None,
        column_name: str = None,
    ):
        """
        Plots a histogram bar-graph.

        Parameters
        ----------
        matrix: Pandas Dataframe
          The dataframe to be plotted.
        low: float, Defaults to 1
          The color mapping value for "low" points.
        high: float, Defaults to 1
          The color mapping value for "high" points.
        title: str, Defaults to None
          The optional title of the heat map.
        tool_tips: list of str, Defaults to None
          An optional list of tool tips to include with the plot.
        column_name: str, Defaults to None
          The name of the column which is being plotted.

        Returns
        -------
        fig: matplotlib Figure
            A matplotlib heatmap figure object.
        """
        mapper = bokeh.models.LinearColorMapper(
            palette=self.bokehpalette, low=low, high=high
        )
        source = bokeh.models.ColumnDataSource(matrix)

        from bokeh.plotting import figure

        p = figure(
            title=f"{title} ({column_name})",
            x_range=(low, high),
            y_range=(0, len(matrix["Y"]) + 1),
            toolbar_location="below",
            toolbar_sticky=False,
            plot_width=600,
            plot_height=600,
        )

        p.hbar(
            y="Y",
            height=0.5,
            left=0,
            source=source,
            right="corr",
            fill_color={"field": "corr", "transform": mapper},
            line_color=None,
        )

        p.add_tools(bokeh.models.HoverTool(tooltips=tool_tips))

        color_bar = bokeh.models.ColorBar(
            color_mapper=mapper,
            major_label_text_font_size="5pt",
            ticker=bokeh.models.BasicTicker(desired_num_ticks=8),
            formatter=bokeh.models.PrintfTickFormatter(format="%0.2f"),
            label_standoff=6,
            border_line_color=None,
            location=(0, 0),
        )

        labels = bokeh.models.LabelSet(
            x="label",
            y="Y",
            text="X",
            level="glyph",
            y_offset=-5,
            source=source,
            render_mode="canvas",
        )

        p.add_layout(color_bar, "right")
        p.add_layout(labels)
        return p

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def generate_heatmap(
        self, corr_matrix, title: str, msg: str, correlation_threshold: float
    ):
        """
        Generate a heatmap from a correlation matrix.


        Parameters
        ----------
        corr_matrix: Pandas Dataframe
          The dataframe to be used for heatmap generation.
        title: str
          title of the heatmap.
        msg: str
          An additional msg to include in the plot.
        correlation_threshold: float
          A float between 0 and 1 which is used for excluding correlations which are not intense enough from the plot.

        Returns
        -------
        tab: matplotlib Panel
            A matplotlib Panel object which includes a plotted heatmap

        """
        from bokeh.plotting import figure

        if len(corr_matrix) == 0:
            tab = bokeh.models.Panel(
                child=figure(title=msg + ", nothing to display"),
                title=title,
            )
            return tab
        corr_matrix = _corr_filter(correlation_threshold, corr_matrix)
        corr_flatten = self.flatten_corr_matrix(corr_matrix)
        low = -1 if corr_matrix.min().min() < 0 else 0
        high = 1
        if self.debug():
            print(f"{title} : Min is {low}, Max is: {high}")

        p = self.plot_heat_map(
            corr_flatten,
            corr_matrix.index.values.tolist(),
            corr_matrix.columns.values.tolist(),
            low,
            high,
            title,
            tool_tips=[("X", "@x"), ("Y", "@y"), ("Corr", "@corr")],
        )

        tab = bokeh.models.Panel(child=p, title=title)
        return tab

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def generate_target_heatmap(
        self,
        corr_matrix,
        title: str,
        correlation_target: str,
        msg: str,
        correlation_threshold: float,
    ):
        """
        Generate a heatmap from a correlation matrix and its targets.


        Parameters
        ----------
        corr_matrix: Pandas Dataframe
          The dataframe to be used for heatmap generation.
        title: str
          title of the heatmap.
        correlation_target: str
          The target column name for computing correlations against.
        msg: str
          An additional msg to include in the plot.
        correlation_threshold: float
          A float between 0 and 1 which is used for excluding correlations which are not intense enough from the plot.

        Returns
        -------
        tab: matplotlib Panel
            A matplotlib Panel object which includes a plotted heatmap.
        """
        from bokeh.plotting import figure

        if len(corr_matrix) == 0:
            tab = bokeh.models.Panel(
                child=figure(title=msg + ", nothing to display"),
                title=title,
            )
            return tab

        corr_matrix = _corr_filter(correlation_threshold, corr_matrix)

        assert correlation_target, "Correlation target is required for this plot"
        if correlation_target not in corr_matrix.columns:
            tab = bokeh.models.Panel(
                child=figure(title="No Data to display"), title=title
            )
            return tab

        corr_flatten = {}

        corr_clean = corr_matrix[correlation_target].dropna()

        if self.debug():
            print(corr_matrix[correlation_target].values.tolist())
            print(len(corr_matrix[correlation_target].values.tolist()))
            print("*" * 30)

        corr_flatten["X"] = corr_clean.index.values.tolist()

        if self.debug():
            print(corr_clean.index.values.tolist())
            print(len(corr_clean.index.values.tolist()))
            print("*" * 30)

        corr_flatten["Y"] = [
            i for i in range(1, len(corr_clean.index.values.tolist()) + 1)
        ]

        corr_flatten["corr"] = corr_clean.values.tolist()
        corr_flatten["label"] = [
            x + 0.01 if x > 0 else 0.01 for x in corr_clean.values.tolist()
        ]

        corr_flatten = pd.DataFrame.from_dict(data=corr_flatten)

        if self.debug():
            print(corr_flatten)

        low = -1 if corr_clean.min() < 0 else 0
        high = 1
        if self.debug():
            print(f"{title} : Min is {low}, Max is: {high}")

        p = self.plot_hbar(
            corr_flatten,
            low,
            high,
            title,
            tool_tips=[("Feature", "@X"), ("Corr", "@corr")],
            column_name=correlation_target,
        )

        tab = bokeh.models.Panel(child=p, title=title)
        return tab

    @runtime_dependency(module="bokeh", install_from=OptionalDependency.VIZ)
    def plot_correlation_heatmap(
        self,
        ds,
        plot_type: str = "heatmap",
        correlation_target: str = None,
        correlation_threshold=-1,
        correlation_methods: str = "pearson",
        **kwargs,
    ):
        """
        Plots a correlation heatmap.

        Parameters
        ----------
        ds: Pandas Slice
          A data slice or file
        plot_type: str Defaults to "heatmap"
          The type of plot - "bar" is another option.
        correlation_target: str, Defaults to None
          the target column for correlation calculations.
        correlation_threshold: float, Defaults to -1
          the threshold for computing correlation heatmap elements.
        correlation_methods: str, Defaults to "pearson"
          the way to compute correlations, other options are "cramers v" and "correlation ratio"
        """
        assert self.ds or ds, "Expecting input for ds or file"

        plot_ds = None

        if ds:
            plot_ds = ds
        else:
            plot_ds = self.ds

        frac = kwargs.get("frac", 1)
        frac = deprecate_default_value(
            frac,
            None,
            1,
            f"<code>frac=None</code> is deprecated. Use <code>frac=1.0</code> instead.",
            FutureWarning,
        )

        force_recompute = kwargs.get("force_recompute", False)
        nan_threshold = kwargs.get("nan_threshold", 0.8)

        cts_cts = pd.DataFrame()
        cat_cat = pd.DataFrame()
        cat_cts = pd.DataFrame()
        correlation_list = plot_ds.corr(
            frac=frac,
            nan_threshold=nan_threshold,
            force_recompute=force_recompute,
            correlation_methods=correlation_methods,
        )

        correlation_methods = _validate_correlation_methods(correlation_methods)
        for method in correlation_methods:
            if method == "pearson":
                cts_cts = plot_ds._pearson
            elif method == "cramers v":
                cat_cat = plot_ds._cramers_v
            elif method == "correlation ratio":
                cat_cts = plot_ds._correlation_ratio
            else:
                raise ValueError(f"This {method} is not supported.")

        # generate the msg
        cts_cts_msg = _generate_msg(
            "pearson", correlation_methods, correlation_target, "continuous"
        )
        cat_cat_msg = _generate_msg(
            "cramers v", correlation_methods, correlation_target, "categorical"
        )
        cat_cts_msg = _generate_msg(
            "correlation ratio",
            correlation_methods,
            correlation_target,
            "continuous and categorical",
        )

        panel_items = [
            {"data": cts_cts, "title": "Continuous vs Continuous", "msg": cts_cts_msg},
            {"data": cat_cat, "title": "Category vs Category", "msg": cat_cat_msg},
            {"data": cat_cts, "title": "Category vs Continuous", "msg": cat_cts_msg},
        ]

        if self.debug():
            print("-" * 50, "Continuous vs Continuous", "-" * 50)
            print(cts_cts.head())

            print("-" * 50, "Category vs Category", "-" * 50)
            print(cat_cat.head())

            print("-" * 50, "Category vs Continuous", "-" * 50)
            print(cat_cts.head())

        if plot_type == "heatmap":
            tabs = [
                self.generate_heatmap(
                    item["data"], item["title"], item["msg"], correlation_threshold
                )
                for item in panel_items
            ]
        elif plot_type == "bar":
            tabs = [
                self.generate_target_heatmap(
                    item["data"],
                    item["title"],
                    correlation_target,
                    item["msg"],
                    correlation_threshold,
                )
                for item in panel_items
            ]
        else:
            raise ValueError("Only supported plot types are heatmap and bar")

        bokeh_tabs = bokeh.models.Tabs(tabs=tabs)

        from bokeh.io import show

        show(bokeh_tabs)


def _generate_msg(
    method: str, correlation_methods: list, correlation_target: str, feature_type: str
):
    if method in correlation_methods or "all" in correlation_methods:
        if correlation_target is not None:
            msg = (
                correlation_target
                + f"is not of the {feature_type} type or not enough of the {feature_type} type(s)"
            )
        else:
            msg = f"Not enough of the {feature_type} type(s)"
    else:
        msg = f'Pass in "{method}" to show the plot'

    return msg


def _has_categorical_features(ds):
    df_type = pd.DataFrame.from_dict(ds.feature_types).T
    return (
        "categorical" in df_type["type"].values or "zipcode" in df_type["type"].values
    )


def plot_correlation_heatmap(ds=None, **kwargs) -> None:
    """
    Plots a correlation heatmap.

    Parameters
    ----------
    ds: Pandas Slice
      A data slice or file
    """
    corr_heatmap_helper = BokehHeatMap(ds)
    corr_heatmap_helper.plot_correlation_heatmap(ds, **kwargs)


def _corr_filter(correlation_threshold, corr):
    if correlation_threshold >= 1 or correlation_threshold < -1:
        raise ValueError(
            "The correlation_threshold value should within the range from -1 to 1."
        )
    new_corr = corr.copy(deep=True)
    np.fill_diagonal(new_corr.values, np.nan)
    new_corr = (
        new_corr[new_corr >= correlation_threshold]
        .dropna(axis=0, how="all")
        .dropna(axis=1, how="all")
    )
    # fill the diagonal with 1 to effectively shrink the correlation matrix
    np.fill_diagonal(new_corr.values, 1)
    return new_corr



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/


from __future__ import print_function, absolute_import, division

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import ContinuousTypedFeature
from ads.common import utils


class ContinuousDetector(AbstractTypeDiscoveryDetector):
    @staticmethod
    def _target_is_continuous(series):
        if str(series.dtype) in ["float16", "float32", "float64"]:
            return True  # treat target variable as continuous
        elif str(series.dtype) in ["int16", "int32", "int64"]:
            if series.nunique() >= 20:
                return True  # treat target variable as continuous

        return False

    def _is_continuous(self, series):
        if series.dtype.name in ["object"]:
            try:
                series.astype("float")
                return True
            except:
                pass

        if series.dtype.name in utils.numeric_pandas_dtypes():
            #
            # if the type is float we simply beleive pandas and go with continuous
            #
            return True

    def discover(self, name, series):

        if self._is_continuous(series):
            logger.debug("column [{}]/[{}] continuous".format(name, series.dtype))
            return ContinuousTypedFeature.build(name, series)

        return False


if __name__ == "__main__":
    dd = ContinuousDetector()
    print(dd.discover("continuous", pd.Series([None, 3.14, 12.0, 1, 2, 3, None])))



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that helps to register custom warnings for the feature types.

Classes
-------
    FeatureWarning
        The Feature Warning class. Provides functionality to register
        warning handlers and invoke them.

Examples
--------
    >>> warning = FeatureWarning()
    >>> def warning_handler_zeros_count(data):
    ...    return pd.DataFrame(
    ...        [['Zeros', 'Age has 38 zeros', 'Count', 38]],
    ...        columns=['Warning', 'Message', 'Metric', 'Value'])
    >>> def warning_handler_zeros_percentage(data):
    ...    return pd.DataFrame(
    ...        [['Zeros', 'Age has 12.2% zeros', 'Percentage', '12.2%']],
    ...        columns=['Warning', 'Message', 'Metric', 'Value'])
    >>> warning.register(name="zeros_count", handler=warning_handler_zeros_count)
    >>> warning.register(name="zeros_percentage", handler=warning_handler_percentage)
    >>> warning.registered()
                        Name                               Handler
        ----------------------------------------------------------
        0         zeros_count          warning_handler_zeros_count
        1    zeros_percentage     warning_handler_zeros_percentage

    >>> warning.zeros_percentage(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros      Age has 38 zeros          Count         38

    >>> warning.zeros_count(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        1          Zeros   Age has 12.2% zeros     Percentage      12.2%

    >>> warning(data_series)
            Warning                    Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros      Age has 38 zeros          Count         38
        1          Zeros   Age has 12.2% zeros     Percentage      12.2%

    >>> warning.unregister('zeros_count')
    >>> warning(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros   Age has 12.2% zeros     Percentage      12.2%
"""
from typing import Callable
import pandas as pd
from ads.feature_engineering.exceptions import WarningNotFound, WarningAlreadyExists


def _validate_warning_handler(handler: Callable) -> bool:
    """Validates warning handler.

    Handler should get pd.Series as a parameter and return pd.DataFrame as result.
    Dataframe should have four columns: Warning, Message, Metric and Value.

    Parameters
    ----------
    handler: Callable
        The handler to validate.

    Returns
    -------
    bool
        True if handler compatible with Feature Warning, False otherwise.
    """
    result = True
    try:
        handler_result = handler(pd.Series([]))
        assert isinstance(handler_result, pd.DataFrame)
        assert list(handler_result.columns) == ["Warning", "Message", "Metric", "Value"]
    except AssertionError:
        result = False
    return result


class FeatureWarning:
    """The Feature Warning class.

    Provides functionality to register warning handlers and invoke them.

    Methods
    -------
    register(self, name: str, handler: Callable) -> None
        Registers a new warning for the feature type.
    unregister(self, name: str) -> None
        Unregisters warning.
    registered(self) -> pd.DataFrame
        Gets the list of registered warnings.

    Examples
    --------
    >>> warning = FeatureWarning()
    >>> def warning_handler_zeros_count(data):
    ...    return pd.DataFrame(
    ...        [['Zeros', 'Age has 38 zeros', 'Count', 38]],
    ...        columns=['Warning', 'Message', 'Metric', 'Value'])
    >>> def warning_handler_zeros_percentage(data):
    ...    return pd.DataFrame(
    ...        [['Zeros', 'Age has 12.2% zeros', 'Percentage', '12.2%']],
    ...        columns=['Warning', 'Message', 'Metric', 'Value'])
    >>> warning.register(name="zeros_count", handler=warning_handler_zeros_count)
    >>> warning.register(name="zeros_percentage", handler=warning_handler_percentage)
    >>> warning.registered()
                      Warning                              Handler
        ----------------------------------------------------------
        0         zeros_count          warning_handler_zeros_count
        1    zeros_percentage     warning_handler_zeros_percentage

    >>> warning.zeros_percentage(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros      Age has 38 zeros          Count         38

    >>> warning.zeros_count(data_series)
                  Warning              Message         Metric      Value
        ----------------------------------------------------------------
        1          Zeros   Age has 12.2% zeros     Percentage      12.2%

    >>> warning(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros      Age has 38 zeros          Count         38
        1          Zeros   Age has 12.2% zeros     Percentage      12.2%

    >>> warning.unregister('zeros_count')
    >>> warning(data_series)
                 Warning               Message         Metric      Value
        ----------------------------------------------------------------
        0          Zeros   Age has 12.2% zeros     Percentage      12.2%
    """

    def __init__(self):
        """Initializes the FeatureWarning."""
        self._data = None
        self._handlers = {}

    def register(self, name: str, handler: Callable, replace: bool = False) -> None:
        """Registers a new warning.

        Parameters
        ----------
        name : str
            The warning name.
        handler: callable
            The handler associated with the warning.
        replace: bool
            The flag indicating if the registered warning should be replaced with the new one.

        Returns
        -------
        None
            Nothing

        Raises
        ------
        ValueError
            If warning name is empty or handler not defined.
        TypeError
            If handler is not callable.
        WarningAlreadyExists
            If warning is already registered.
        """
        if not name:
            raise ValueError("Warning name is not provided.")
        if name in self._handlers and not replace:
            raise WarningAlreadyExists(name)
        if not handler:
            raise ValueError("Handler is not provided.")
        if not callable(handler):
            raise TypeError("Handler should be a function.")
        self._handlers[name] = handler

    def unregister(self, name: str) -> None:
        """Unregisters warning.

        Parameters
        -----------
        name: str
            The name of warning to be unregistered.

        Returns
        -------
        None
            Nothing.

        Raises
        ------
        ValueError
            If warning name is not provided or empty.
        WarningNotFound
            If warning not found.
        """
        if not name:
            raise ValueError("Warning name is not provided.")
        if name not in self._handlers:
            raise WarningNotFound(name)
        del self._handlers[name]

    def registered(self) -> pd.DataFrame:
        """Gets the list of registered warnings.

        Returns
        -------
        pd.DataFrame

        Examples
        --------
        >>>    The list of registerd warnings in DataFrame format.
                             Name                               Handler
            -----------------------------------------------------------
            0         zeros_count           warning_handler_zeros_count
            1    zeros_percentage      warning_handler_zeros_percentage
        """
        result = []
        for name, handler in self._handlers.items():
            result.append((name, handler.__name__))
        return pd.DataFrame(result, columns=["Warning", "Handler"])

    def _bind_data(self, data: pd.Series) -> None:
        """Binds data to the feature warning.

        Parameters
        ----------
        data: pd.Series
            The data to be bound.
        """
        self._data = data

    def _process(self) -> pd.DataFrame:
        """Invokes the all registered warnings.

        Returns
        -------
        pd.DataFrame
        >>>    The result of invoked warning handlers.
                 Warning               Message       Metric    Value
            --------------------------------------------------------
                   Zeros      Age has 38 zeros        Count       38
                   Zeros   Age has 12.2% zeros   Percentage    12.2%
        Raises
        ------
        ValueError
            If data is not provided or result of warning has a wrong format.
        """
        if self._data is None:
            raise ValueError("Data is not provided.")

        if not self._handlers:
            return None

        expected_columns = ["Warning", "Message", "Metric", "Value"]
        result_df = pd.DataFrame([], columns=expected_columns)
        for name, handler in self._handlers.items():
            try:
                handler_result = handler(self._data)
            except Exception as ex:
                raise ValueError(
                    f"An error occurred while executing the '{name}'. "
                    f"Details: {str(ex)}."
                ) from ex
            if handler_result is not None:
                if not isinstance(handler_result, pd.DataFrame) or (
                    not handler_result.empty
                    and set(list(handler_result.columns)) != set(expected_columns)
                ):
                    raise ValueError(
                        f"An error occurred while executing the '{name}'. "
                        f"Details: '{name}' should return a DataFrame "
                        f"with columns: {expected_columns}."
                    )
                result_df = result_df.append(handler_result)
        result_df.reset_index(drop=True, inplace=True)
        return result_df

    def __call__(self, *args) -> pd.DataFrame:
        """Makes class instance callable.

        Parameters
        ----------
        *args
            Variable length argument list.

        Returns
        -------
        pd.DataFrame
            The result of processing the all registered warnings.

        Raises
        ------
        ValueError
            If data is not provided,
        TypeError
            If data has wrong format.
        """
        if args and len(args) > 0:
            self._data = args[0]
        if self._data is None:
            raise ValueError("Data is not provided.")
        if not isinstance(self._data, pd.Series):
            raise TypeError("Wrong data format. Data should be Series.")
        return self._process()

    def __getattr__(self, attr):
        """Makes it possible to invoke registered warning as a regular method."""
        if attr in self._handlers:
            return self._handlers[attr]
        raise AttributeError(attr)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import, division

import pandas as pd

from ads.type_discovery import logger
from ads.type_discovery.abstract_detector import AbstractTypeDiscoveryDetector
from ads.type_discovery.typed_feature import ConstantTypedFeature


class ConstantDetector(AbstractTypeDiscoveryDetector):
    def is_constant(self, name, values):
        #
        # if all the values are null we treat this as a const feature
        #

        return values.size == 0 or values.nunique() == 1

    def discover(self, name, series):
        candidates = series.loc[~series.isnull()]

        if self.is_constant(name, candidates):
            logger.debug("column [{}]/[{}] Constant".format(name, series.dtype))
            return ConstantTypedFeature.build(name, series)

        return False


if __name__ == "__main__":
    dd = ConstantDetector()
    print(dd.discover("zipcodes", pd.Series([None, "94065", "94065", "94065", None])))
    print(dd.discover("years", pd.Series([2008, 2008, 2008, 2008, 2008])))
    df = pd.read_csv(
        "https://artifacthub.oraclecorp.com/dsc-generic/advanced-ds/datasets/flights.csv"
    )
    print(dd.discover("flights_years", df["Year"]))



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import abc


class AbstractTypeDiscoveryDetector(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def discover(self, name, series):
        return


class DiscreteDiscoveryDetector(AbstractTypeDiscoveryDetector, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def discover(self, name, series):
        return



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import os
from typing import Dict, Generator, List, Union

from ads.text_dataset import backends
from ads.text_dataset.backends import OITCC, Base, PDFPlumber, Tika
from ads.text_dataset.utils import NotSupportedError
from fsspec.core import OpenFile

logger = logging.getLogger("ads.text_dataset")


class FileProcessor:
    """
    Base class for all the file processor. Files are opened using fsspec library.
    The default implementation in the base class assumes text files.

    This class is expected to be used inside `ads.text_dataset.dataset.DataLoader`.
    """

    backend_map = {"default": Base, "tika": Tika}

    def __init__(self, backend: Union[str, backends.Base] = "default") -> None:
        self.backend(backend)

    def backend(self, backend: Union[str, backends.Base]) -> None:
        """Set backend for file processor.

        Parameters
        ----------
        backend : `ads.text_dataset.backends.Base`
            a backend for file processor

        Returns
        -------
        None

        Raises
        ------
        NotSupportedError
            when specified backend is not supported.
        """
        if isinstance(backend, str) and backend in self.backend_map:
            self._backend = self.backend_map[backend]()
        elif isinstance(backend, Base):
            self._backend = backend
        else:
            raise NotSupportedError(
                f"backend {backend} is not recognized or not a subclass of ads.text_dataset.backends.Base."
            )
        return self

    def read_line(
        self, fhandler: OpenFile, **format_reader_kwargs: Dict
    ) -> Generator[Union[str, List[str]], None, None]:
        """Yields lines from a file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            file handler returned by `fsspec`

        Returns
        -------
        Generator
            a generator that yields lines from a file
        """
        return self._backend.read_line(fhandler, **format_reader_kwargs)

    def read_text(
        self, fhandler: OpenFile, **format_reader_kwargs: Dict
    ) -> Generator[Union[str, List[str]], None, None]:
        """Yield contents from the entire file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by fsspec

        Returns
        -------
        Generator
            a generator that yield text from a file
        """
        return self._backend.read_text(fhandler, **format_reader_kwargs)

    def convert_to_text(
        self,
        fhandler: OpenFile,
        dst_path: str,
        fname: str = None,
        storage_options: Dict = None,
    ) -> str:
        """Convert input file to a text file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by `fsspec`
        dst_path: str
            local folder or cloud storage (e.g. OCI object storage) prefix to save converted text files
        fname: str, optional
            filename for converted output, relative to dirname or prefix, by default None
        storage_options: dict, optional
            storage options for cloud storage, by default None

        Returns
        -------
        str
            path to saved output
        """
        return self._backend.convert_to_text(fhandler, dst_path, fname, storage_options)

    def get_metadata(self, fhandler: OpenFile) -> Dict:
        """Get metadata of a file.

        Parameters
        ----------
        fhandler : `fsspec.core.OpenFile`
            a file handler returned by fsspec

        Returns
        -------
        dict
            dictionary of metadata
        """
        return self._backend.get_metadata(fhandler)


class PDFProcessor(FileProcessor):
    """
    Extracts text content from PDF
    """

    backend_map = {"tika": Tika, "pdfplumber": PDFPlumber, "default": Tika}


class WordProcessor(FileProcessor):
    """
    Extracts text content from doc or docx format.
    """

    backend_map = {"default": Tika, "tika": Tika}


class FileProcessorFactory:
    """Factory that manages all file processors.
    Provides functionality to get a processor corresponding to a given file type,
    or register custom processor for a specific file format.

    Examples
    --------
    >>> from ads.text_dataset.extractor import FileProcessor, FileProcessorFactory
    >>> FileProcessorFactory.get_processor('pdf')
    >>> class CustomProcessor(FileProcessor):
    ... # custom logic here
    ... pass
    >>> FileProcessorFactory.register('new_format', CustomProcessor)
    """

    processor_map = {
        "pdf": PDFProcessor,
        "docx": WordProcessor,
        "doc": WordProcessor,
        "txt": FileProcessor,
    }

    @classmethod
    def register(cls, fmt: str, processor: FileProcessor) -> None:
        """Register custom file processor for a file format.

        Parameters
        ----------
        fmt : str
            file format
        processor : `FileProcessor`
            custom processor

        Raises
        ------
        TypeError
            raised when processor is not a subclass of `FileProcessor`.
        """
        if issubclass(processor, FileProcessor):
            cls.processor_map[fmt] = processor
        else:
            raise TypeError(f"Processor must inherit from FileProcessor class.")

    @staticmethod
    def get_processor(format):
        if format in FileProcessorFactory.processor_map:
            return FileProcessorFactory.processor_map[format]
        else:
            logger.warning(
                f"""
                Format {format} is not supported natively.
                A generic FileProcessor is returned.
                You can define and register a custom processor.
                """
            )
            return FileProcessor



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import ads
from ads.secrets import SecretKeeper, Secret
import json
import os
import tempfile
import zipfile
from tqdm.auto import tqdm

logger = ads.getLogger("ads.secrets")

from dataclasses import dataclass, field


@dataclass
class ADBSecret(Secret):
    """
    Dataclass representing the attributes managed and serialized by ADBSecretKeeper
    """

    user_name: str
    password: str
    service_name: str
    wallet_location: str = field(
        default=None, metadata={"serializable": False}
    )  # Not saved in vault
    wallet_file_name: str = field(
        default=None, repr=False
    )  # Not exposed through environment or `to_dict` function
    wallet_content: dict = field(
        default=None,
        repr=False,
        metadata={"serializable": False},  # Not saved in vault and not saved in vault
    )
    wallet_secret_ids: list = field(
        repr=False, default_factory=list
    )  # Not exposed through environment or `to_dict` function

    def __post_init__(self):
        self.wallet_file_name = (
            os.path.basename(self.wallet_location)
            if self.wallet_location
            else self.wallet_file_name
        )


class ADBSecretKeeper(SecretKeeper):
    """
    `ADBSecretKeeper` provides an interface to save ADW/ATP database credentials.
    This interface does not store the wallet file by default.
    For saving wallet file, set `save_wallet=True` while calling `ADBSecretKeeper.save` method.


    Examples
    --------
    >>> # Saving credentials without saving the wallet file
    >>> from ads.secrets.adw import ADBSecretKeeper
    >>> vault_id = "ocid1.vault.oc1..<unique_ID>"
    >>> key_id = "ocid1.key..<unique_ID>"

    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> connection_parameters={
    ...     "user_name":"admin",
    ...     "password":"<your password>",
    ...     "service_name":"service_name_{high|low|med}",
    ...     "wallet_location":"/home/datascience/Wallet_xxxx.zip"
    ... }
    >>> adw_keeper = ADBSecretKeeper(vault_id=vault_id, key_id=key_id, **connection_parameters)
    >>> adw_keeper.save("adw_employee", "My DB credentials", freeform_tags={"schema":"emp"}) # Does not save the wallet file
    >>> print(adw_keeper.secret_id) # Prints the secret_id of the stored credentials
    >>> adw_keeper.export_vault_details("adw_employee_att.json", format="json") # Save the secret id and vault info to a json file

    >>> # Loading credentails
    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> from ads.secrets.adw import ADBSecretKeeper
    >>> secret_id = "ocid1.vaultsecret.oc1..<unique_ID>"
    >>> with ADBSecretKeeper.load_secret(source=secret_id,
                                 wallet_location='/home/datascience/Wallet_xxxxxx.zip') as adw_creds:
    ...     import pandas as pd
    ...     df = pd.DataFrame.ads.read_sql("select * from EMPLOYEE", connection_parameters=adw_creds)


    >>> myadw_creds = ADBSecretKeeper.load_secret(source='adw_employee_att.json', format="json"
    ...                          wallet_location='/home/datascience/Wallet_xxxxxx.zip')
    >>> pd.DataFrame.ads.read_sql("select * from ATTRITION_DATA", connection_parameters=myadw_creds.to_dict()).head(2)

    >>> # Saving and loading credentials with wallet storage
    >>> # Saving credentials
    >>> from ads.secrets.adw import ADBSecretKeeper
    >>> vault_id = "ocid1.vault.oc1..<unique_ID>"
    >>> key_id = "ocid1.key.oc1..<unique_ID>"

    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> connection_parameters={
    ...     "user_name":"admin",
    ...     "password":"<your password>",
    ...     "service_name":"service_name_{high|low|med}",
    ...     "wallet_location":"/home/datascience/Wallet_xxxx.zip"
    ... }
    >>> adw_keeper = ADBSecretKeeper(vault_id=vault_id, key_id=key_id, **connection_parameters)
    >>> adw_keeper.save("adw_employee", "My DB credentials", freeform_tags={"schema":"emp"}, save_wallet=True)
    >>> print(adw_keeper.secret_id) # Prints the secret_id of the stored credentials
    >>> adw_keeper.export_vault_details("adw_employee_att.json") # Save the secret id and vault info to a json file

    >>> # Loading credentails
    >>> import ads
    >>> ads.set_auth("resource_principal") # If using resource principal for authentication
    >>> from ads.secrets.adw import ADBSecretKeeper
    >>> secret_id = "ocid1.vaultsecret.oc1..<unique_ID>"
    >>> with ADBSecretKeeper.load_secret(source=secret_id) as adw_creds:
    ...     import pandas as pd
    ...     df = pd.DataFrame.ads.read_sql("select * from EMPLOYEE", connection_parameters=adw_creds)


    >>> myadw_creds = ADBSecretKeeper.load_secret(source='adw_employee_att.json', format='json')
    >>> pd.DataFrame.ads.read_sql("select * from ATTRITION_DATA", connection_parameters=myadw_creds.to_dict()).head(2)
    """

    def __init__(
        self,
        user_name: str = None,
        password: str = None,
        service_name: str = None,
        wallet_location: str = None,
        wallet_dir: str = None,
        repository_path: str = None,
        repository_key: str = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        user_name: (str, optioanl). Default None
            user_name of the databse
        password: (str, optional). Default None
            password  for connecting to the database
        service_name: (str, optional). Default None
            service name of the ADB instance
        wallet_location: (str, optional). Default None
            full path to the wallet zip file used for connecting to ADB instance.
        wallet_dir: (str, optional). Default None
            local directory where the extracted wallet content is saved
        repository_path: (str, optional). Default None.
            Path to credentials repository. For more details refer `ads.database.connection`
        repository_key: (str, optional). Default None.
            Configuration key for loading the right configuration from repository. For more details refer `ads.database.connection`
        kwargs:
            vault_id: str. OCID of the vault where the secret is stored. Required for saving secret.
            key_id: str. OCID of the key used for encrypting the secret. Required for saving secret.
            compartment_id: str. OCID of the compartment where the vault is located. Required for saving secret.
            auth: dict. Dictionay returned from ads.common.auth.api_keys() or ads.common.auth.resource_principal(). By default, will follow what is set in `ads.set_auth`.  Use this attribute to override the default.
        """
        if repository_path and repository_key:
            from ads.database import connection

            config_from_repo = connection.get_repository(
                repository_key, repository_path
            )
            if user_name is None:
                user_name = config_from_repo.get("username") or config_from_repo.get(
                    "user_name"
                )
            if password is None:
                password = config_from_repo.get("password")
            if service_name is None:
                service_name = config_from_repo.get("service_name")
            if wallet_location is None:
                wallet_location = config_from_repo.get("wallet_location")

        self.data = ADBSecret(
            user_name=user_name,
            password=password,
            service_name=service_name,
            wallet_location=wallet_location,
        )
        self.wallet_dir = wallet_dir

        super().__init__(**kwargs)

    def encode(
        self, serialize_wallet: bool = False
    ) -> "ads.secrets.adb.ADBSecretKeeper":
        """
        Prepares content to save in vault. The user_name, password and service_name and the individual files inside the wallet zip file are base64 encoded and stored in `self.secret`

        Parameters
        ----------
        serialize_wallet: bool, optional
            When set to True, loads the wallet zip file and encodes the content of each file in the zip file.

        Returns
        -------
        ADBSecretKeeper:
            Returns self object
        """

        if serialize_wallet:
            if not self.data.wallet_location:
                raise ValueError(
                    "Missing path to wallet zip file. Required wallet zip file path to be set in `wallet_location` "
                )
            with tempfile.TemporaryDirectory() as tmpdir:
                zipfile.ZipFile(self.data.wallet_location).extractall(tmpdir)
                wallet_info = {}
                with tqdm(os.listdir(tmpdir), leave=False) as pbar:
                    for file in pbar:
                        pbar.set_description(f"Encoding: {file}")
                        with open(os.path.join(tmpdir, file), "rb") as f:
                            encoded = self._encode(f.read())
                            wallet_info[file] = encoded
                self.wallet_content = wallet_info
                self.secret = {
                    **self.data.serialize(),
                    "wallet_content": self.wallet_content,
                }
                self.encoded = self.secret
        else:
            self.data.wallet_file_name = None
            super().encode()
        return self

    def decode(self) -> "ads.secrets.adb.ADBSecretKeeper":
        """
        Converts the content in `self.secret` to `ADBSecret` and stores in `self.data`

        If the `wallet_location` is passed through the constructor, then retain it. We do not want to override what user has passed in
        If the `wallet_location` was not passed, but the sercret has `wallet_secret_ids`, then we generate the wallet zip file in the location specified by `wallet_dir` in the constructor

        Returns
        -------
        ADBSecretKeeper:
            Returns self object
        """
        content = json.loads(self._decode())
        logger.debug(f"Decoded secret contains following keys - {list(content.keys())}")
        data = ADBSecret(
            **content,
        )

        if hasattr(self, "data") and self.data.wallet_location:
            # If the wallet location is passed in as a keyword argument inside
            # `load_secret` method, then we need to retain that information after
            # the secret was deserialized.
            logger.debug(f"Setting wallet file to {self.data.wallet_location}")
            data.wallet_location = self.data.wallet_location
        elif data.wallet_secret_ids and len(data.wallet_secret_ids) > 0:
            logger.debug(f"Secret ids corresponding to the wallet files found.")
            # If the secret ids for wallet files are available in secret, then we
            # can generate the wallet file.

            if self.wallet_dir:
                os.makedirs(self.wallet_dir, exist_ok=True)
            data.wallet_location = (
                os.path.join(self.wallet_dir or "", data.wallet_file_name)
                if data.wallet_file_name
                else None
            )
            with tempfile.TemporaryDirectory() as tmpdir:
                wallet_files = []
                with tqdm(data.wallet_secret_ids, leave=False) as pbar:
                    for wsec_id in pbar:
                        pbar.set_description(f"Fetching secret content for {wsec_id}")
                        content = json.loads(
                            self._decode(self.get_secret(wsec_id, decoded=False))
                        )
                        wallet_files.append(os.path.join(tmpdir, content["filename"]))
                        with open(
                            os.path.join(tmpdir, content["filename"]), "wb"
                        ) as wf:
                            wf.write(
                                self._decode(content["content"], str_encoding=None)
                            )
                logger.debug(f"Creating wallet zip file at {data.wallet_location}")
                with zipfile.ZipFile(data.wallet_location, "w") as wzipf:
                    for file in wallet_files:
                        wzipf.write(file, os.path.basename(file))
                logger.debug(
                    f"Wallet zip file create successully at {data.wallet_location}"
                )
        else:
            logger.info(
                "Wallet information not found. If you have a local wallet zip file, you can pass that as a keyword argument `wallet_location=<path to the wallet zip file>` inside the `ADBSecretKeeper.load_secret` method"
            )
        self.data = data
        return self

    def save(
        self,
        name: str,
        description: str,
        freeform_tags: dict = None,
        defined_tags: dict = None,
        save_wallet: bool = False,
    ) -> "ads.secrets.adb.ADBSecretKeeper":
        """Saves credentials to Vault and returns self.

        Parameters
        ----------
        name : str
            Name of the secret when saved in the Vault.
        description : str
            Description of the secret when saved in the Vault.
        freeform_tags : (dict, optional). Default is None
            freeform_tags to be used for saving the secret in OCI console.
        defined_tags: (dict, optional). Default is None
            Save the tags under predefined tags in OCI console.
        save_wallet: (bool, optional). Default is False
                If set to True, saves the contents of the wallet file as separate secret.

        Returns
        -------
        ADBSecretKeeper:
            Returns self object
        """
        if not save_wallet:
            logger.info(
                "`save_wallet` set to False. Not saving wallet file content to vault"
            )
        logger.debug(f"Encoding secrets. Save wallet is set to {save_wallet}")
        self.encode(serialize_wallet=save_wallet)
        secret_id_list = []
        if save_wallet:
            with tqdm(self.encoded["wallet_content"], leave=False) as pbar:
                for file in pbar:
                    pbar.set_description(f"Saving {file}")
                    secret_id = self.create_secret(
                        self._encode(
                            json.dumps(
                                {
                                    "filename": file,
                                    "content": self.encoded["wallet_content"][file],
                                }
                            ).encode("utf-8")
                        ),
                        encode=False,
                        secret_name=f"{name}_{file}",
                        description=f"{description}\n {file}",
                        freeform_tags=freeform_tags,
                        defined_tags=defined_tags,
                    )
                    secret_id_list.append(secret_id)

        self.data.wallet_secret_ids = secret_id_list
        with tqdm(
            [1], desc="Saving credentials and file secret ids", leave=False
        ) as pbar:
            for _ in pbar:
                self.secret_id = self.create_secret(
                    self._encode(json.dumps(self.data.serialize()).encode("utf-8")),
                    encode=False,
                    secret_name=name,
                    description=description,
                    freeform_tags=freeform_tags,
                    defined_tags=defined_tags,
                )
        return self



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import abc
import ads.common.utils as utils
from oci.util import to_dict
from pandas import DataFrame
import pandas as pd
from abc import ABCMeta
from ads.common.decorator.deprecate import deprecated


class SummaryList(list, metaclass=ABCMeta):
    @deprecated("2.6.3")
    def __init__(self, entity_list, datetime_format=utils.date_format):
        if isinstance(entity_list, filter):
            entity_list = list(entity_list)
        super(SummaryList, self).__init__(entity_list)

        df = DataFrame.from_records([to_dict(entity) for entity in entity_list])

        # edge case handling: handle edge case where dataframe is empty
        if not df.empty:
            # handle collision protection
            nuniques = df["id"].nunique()
            minLen = -6
            shortened_ocids = df["id"].str[minLen:]
            while shortened_ocids.nunique() != nuniques:
                minLen -= 1
                shortened_ocids = df["id"].str[minLen:]
            df.index = shortened_ocids
            self.short_id_index = df["id"].to_dict()
            ordered_columns = [
                "display_name",
                "time_created",
                "lifecycle_state",
            ]

            ordered_columns.extend(
                [column for column in df.columns if column not in ordered_columns]
            )

            for column in ordered_columns:
                if column not in df.columns:
                    ordered_columns.remove(column)
            self.df = df[ordered_columns].drop(columns="id")
            self.df["time_created"] = pd.to_datetime(
                self.df["time_created"]
            ).dt.strftime(datetime_format)
            self.datetime_format = datetime_format
        else:
            self.df = df

    @abc.abstractmethod
    def sort_by(self, columns, reverse=False):  # pragma: no cover
        """
        Abstract sort method for dataflow summary.
        """
        pass

    @abc.abstractmethod
    def filter(self, selection, instance=None):  # pragma: no cover
        """
        Abstract filter method for dataflow summary.
        """
        pass

    @abc.abstractmethod
    def __add__(self, rhs):  # pragma: no cover
        pass

    def to_dataframe(self, datetime_format=None):
        """
        Abstract to_dataframe method for dataflow summary.
        """
        df = self.df.copy()
        if datetime_format is not None:
            df["time_created"] = pd.to_datetime(df["time_created"]).dt.strftime(
                datetime_format
            )
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that represents a Phone Number feature type.

Classes:
    PhoneNumber
        The Phone Number feature type.

Functions:
    default_handler(data: pd.Series) -> pd.Series
        Processes given data and indicates if the data matches requirements.
"""
import pandas as pd
import re
from ads.feature_engineering.feature_type.string import String
from ads.feature_engineering.utils import _count_unique_missing
from ads.feature_engineering import schema


PATTERN = re.compile(
    r"^(\+?\d{1,2}[\s-])?\(?(\d{3})\)?[\s.-]?\d{3}[\s.-]?\d{4}$", re.VERBOSE
)


def default_handler(data: pd.Series, *args, **kwargs) -> pd.Series:
    """Processes given data and indicates if the data matches requirements.

    Parameters
    ----------
    data: :class:`pandas.Series`
        The data to process.

    Returns
    -------
    :class:`pandas.Series`
        The logical list indicating if the data matches requirements.
    """

    def _is_phone_number(x: any):
        return (
            not pd.isnull(x) and isinstance(x, str) and re.match(PATTERN, x) is not None
        )

    return data.apply(lambda x: True if _is_phone_number(x) else False)


class PhoneNumber(String):
    """Type representing phone numbers.

    Attributes
    -----------
    description: str
        The feature type description.
    name: str
        The feature type name.
    warning: FeatureWarning
        Provides functionality to register warnings and invoke them.
    validator
        Provides functionality to register validators and invoke them.

    Methods
    --------
    feature_stat(x: pd.Series) -> pd.DataFrame
        Generates feature statistics.

    Examples
    --------
    >>> from ads.feature_engineering.feature_type.phone_number import PhoneNumber
    >>> import pandas as pd
    >>> s = pd.Series([None, "1-640-124-5367", "1-573-916-4412"])
    >>> PhoneNumber.validator.is_phone_number(s)
    0    False
    1     True
    2     True
    dtype: bool
    """

    description = "Type representing phone numbers."

    @staticmethod
    def feature_stat(x: pd.Series) -> pd.DataFrame:
        """Generates feature statistics.

        Feature statistics include (total)count, unique(count) and
        missing(count) if there is any.

        Examples
        --------
        >>> s = pd.Series(['2068866666', '6508866666', '2068866666', '', np.NaN, np.nan, None], name='phone')
        >>> s.ads.feature_type = ['phone_number']
        >>> s.ads.feature_stat()
            Metric  Value
        1	count	7
        2	unique	2
        3	missing	4

        Returns
        -------
        :class:`pandas.DataFrame`
            Summary statistics of the Series or Dataframe provided.
        """
        return _count_unique_missing(x)

    @classmethod
    def feature_domain(cls, x: pd.Series) -> schema.Domain:
        """
        Generate the domain of the data of this feature type.

        Examples
        --------
        >>> s = pd.Series(['2068866666', '6508866666', '2068866666', '', np.NaN, np.nan, None], name='phone')
        >>> s.ads.feature_type = ['phone_number']
        >>> s.ads.feature_domain()
        constraints: []
        stats:
            count: 7
            missing: 4
            unique: 2
        values: PhoneNumber

        Returns
        -------
        ads.feature_engineering.schema.Domain
            Domain based on the PhoneNumber feature type.
        """

        return schema.Domain(cls.__name__, cls.feature_stat(x).to_dict()[x.name], [])


PhoneNumber.validator.register("is_phone_number", default_handler)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from io import BytesIO
from typing import Any, Dict, List, Union

import fsspec
import PIL
from ads.common import auth as authutil
from ads.data_labeling.constants import DatasetType
from ads.data_labeling.interface.loader import Loader
from ads.text_dataset.dataset import TextDatasetFactory, backends
from PIL import Image

THREAD_POOL_MAX_WORKERS = 10


class FileLoader:
    """FileLoader Base Class.

    Attributes:
    ----------
    auth: (dict, optional). Defaults to None.
        The default authetication is set using `ads.set_auth` API. If you need to override the
        default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
        authentication signer and kwargs required to instantiate IdentityClient object.

    Examples
    --------
    >>> from ads.data_labeling.loader.file_loader import FileLoader
    >>> import oci
    >>> import os
    >>> from ads.common import auth as authutil
    >>> path = "path/to/your_text_file.txt"
    >>> file_content = FileLoader(auth=authutil.api_keys()).load(path)
    """

    def __init__(self, auth: Dict = None) -> "FileLoader":
        """Initiates a FileLoader instance.

        Parameters
        ----------
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        """
        self.auth = auth or authutil.default_signer()

    def load(self, path: str, **kwargs) -> BytesIO:
        """Loads the file content from the path.

        Parameters
        ----------
        path: str
            The file path, can be local or object storage path.
        kwargs:
            Nothing.

        Returns
        -------
        BytesIO
            The data in BytesIO format.
        """
        data = None
        try:
            with fsspec.open(path, **self.auth) as f:
                data = f.read()
        except FileNotFoundError:
            raise FileNotFoundError(f"{path} not found.")
        except Exception as e:
            raise e

        return data

    def bulk_load(self, paths: List[str], **kwargs) -> Dict[str, Any]:
        """Loads the files content from the list of paths.
        The ThreadPoolExecutor is used to load the files in parallel threads.

        Parameters
        ----------
        paths: List[str]
            The list of file paths, can be local or object storage paths.

        Returns
        -------
        Dict[str, Any]
            The map between file path and file content.
        """
        result = {}
        if not paths or not isinstance(paths, list) or len(paths) == 0:
            return result

        with ThreadPoolExecutor(max_workers=THREAD_POOL_MAX_WORKERS) as pool:
            futures = {pool.submit(self.load, path, **kwargs): path for path in paths}
            result = {futures[task]: task.result() for task in as_completed(futures)}
        return result


class TextFileLoader(FileLoader):
    """
    TextFileLoader class which loads text files.

    Examples
    --------
    >>> from ads.data_labeling import TextFileLoader
    >>> import oci
    >>> import os
    >>> from ads.common import auth as authutil
    >>> path = "path/to/your_text_file.txt"
    >>> file_content = TextFileLoader(auth=authutil.api_keys()).load(path)
    """

    def load(
        self, path: str, backend: Union[str, backends.Base] = "default", **kwargs
    ) -> str:
        """Loads the content from the path.

        Parameters
        ----------
        path: str
            Text file path, can be local or object storage path.
        backend: Union[str, backends.Base]
            Default to "default". Valid options are "default" and "tika" or
            ads.text_dataset.backends.Base, ads.text_dataset.backends.Tika
        kwargs:
            encoding: (str, optional). Defaults to 'utf-8'.
                Encoding for text files. Used only to extract the content of the text dataset contents.

        Returns
        -------
        str
            Content of the text file.
        """
        format = os.path.splitext(path)[1].replace(".", "")
        try:
            content = next(
                TextDatasetFactory.format(format.lower())
                .backend(backend)
                .read_text(path, storage_options=self.auth, **kwargs)
            )[0]
        except FileNotFoundError:
            raise FileNotFoundError(f"{path} not found.")
        except Exception as e:
            raise e
        return content


class ImageFileLoader(FileLoader):
    """
    ImageFileLoader class which loads image files.

    Examples
    --------
    >>> from ads.data_labeling import ImageFileLoader
    >>> import oci
    >>> import os
    >>> from ads.common import auth as authutil
    >>> path = "path/to/image.png"
    >>> image = ImageFileLoader(auth=authutil.api_keys()).load(path)
    """

    def load(self, path: str, **kwargs) -> PIL.ImageFile.ImageFile:
        """Loads the image from the path.

        Parameters
        ----------
        path: str
            Image file path, can be local or object storage path.
        kwargs:
            Nothing.

        Returns
        -------
        PIL.ImageFile.ImageFile
            Image opened by Pillow.
        """
        data = None
        data = super().load(path=path)

        return Image.open(BytesIO(data))


class FileLoaderFactory:
    """FileLoaderFactory class to create/register FileLoaders."""

    _loaders = {
        DatasetType.TEXT: TextFileLoader,
        DatasetType.IMAGE: ImageFileLoader,
        DatasetType.DOCUMENT: FileLoader,
    }

    @staticmethod
    def loader(dataset_type: str, auth: Dict = None) -> FileLoader:
        """Gets the loader based on the dataset_type.

        Parameters
        ----------
        dataset_type: str
            Dataset type. Currently supports TEXT, IMAGE and DOCUMENT.
        auth: (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        FileLoader
            A FileLoader instance corresponding to the dataset_type.
        """
        if not dataset_type in FileLoaderFactory._loaders:
            raise ValueError(
                f"The wrong dataset type has been provided. "
                f"Supported dataset types are: `{DatasetType.TEXT}`, "
                f"`{DatasetType.IMAGE}` and `{DatasetType.DOCUMENT}`."
            )

        return FileLoaderFactory._loaders[dataset_type](auth=auth)

    @classmethod
    def register(cls, dataset_type: str, loader: Loader) -> None:
        """Registers a new loader for a given dataset_type.

        Parameters
        ----------
        dataset_type: str
            Dataset type. Currently supports TEXT and IMAGE.
        loader: Loader
            A Loader class which supports loading content of the given dataset_type.

        Returns
        -------
        None
            Nothing.
        """
        cls._parsers[dataset_type] = loader



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
from datetime import datetime
from types import MethodType
import time
import pathlib
import json
import re
import uuid
from urllib.parse import urlparse
from pathlib import Path
from pandas import DataFrame
from ast import literal_eval

from jinja2 import Environment, PackageLoader

from ads.common import oci_client, auth, logger
from ads.common.utils import FileOverwriteError
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.decorator.deprecate import deprecated

from oci.data_flow.models import (
    CreateApplicationDetails,
    Application,
    ApplicationParameter,
    ApplicationSummary,
    UpdateApplicationDetails,
    CreateRunDetails,
    Run,
    UpdateRunDetails,
    RunSummary,
)
from oci.exceptions import ServiceError

from ads.common import utils
from ads.dataflow.dataflowsummary import SummaryList
from ads.config import NB_SESSION_COMPARTMENT_OCID, TENANCY_OCID, OCI_REGION_METADATA

env = Environment(loader=PackageLoader("ads", "templates"))

create_application_details_attributes = CreateApplicationDetails().swagger_types.keys()
update_application_details_attributes = UpdateApplicationDetails().swagger_types.keys()
application_attributes = list(Application().swagger_types.keys())

create_run_details_attributes = CreateRunDetails().swagger_types.keys()
update_run_details_attributes = UpdateRunDetails().swagger_types.keys()
run_attributes = list(Run().swagger_types.keys())


class SPARK_VERSION(str):
    v2_4_4 = "2.4.4"
    v3_0_2 = "3.0.2"


class DataFlow:
    @deprecated(
        "2.6.3",
        details="Use ads.jobs.DataFlow class for creating DataFlow applications and runs. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/apachespark/dataflow.html#create-run-data-flow-application-using-ads-python-sdk",
    )
    def __init__(
        self,
        compartment_id=None,
        dataflow_base_folder="/home/datascience/dataflow",
        os_auth=None,
        df_auth=None,
    ):
        # create iff not found dataflow_base_folder
        self.dataflow_base_folder = dataflow_base_folder
        self.os_auth = os_auth if os_auth else auth.default_signer()
        self.df_auth = df_auth if df_auth else auth.default_signer()
        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )
        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")
        self.display_name = None
        self.driver_shape = None
        self.executor_shape = None
        self.file_uri = None
        self.archive_uri = None
        self.language = None
        self.logs_bucket_uri = None
        self.num_executors = None
        self.spark_version = None
        self.warehouse_bucket_uri = None

        self.object_storage_client = oci_client.OCIClientFactory(
            **self.os_auth
        ).object_storage
        self.df_client = oci_client.OCIClientFactory(**self.df_auth).dataflow

        self.region = (
            self.df_auth["config"]["region"]
            if "config" in self.df_auth and "region" in self.df_auth["config"]
            else (
                literal_eval(OCI_REGION_METADATA)["regionIdentifier"]
                if OCI_REGION_METADATA
                else None
            )
        )
        if not self.region:
            logger.warning(
                "Region information not found from oci config file. Set region in the OCI config file"
            )
        try:
            self.namespace = self.object_storage_client.get_namespace().data
        except ServiceError as se:
            if se.status == 404:
                raise ValueError(
                    f'The compartment_id "{self.compartment_id}" have to be '
                    f"in same tenancy as current user "
                ) from se
            else:
                raise

        self.short_id_index = {}
        # Currently Data Flow only supports VM.Standard.2 series
        # VM_shapes dict needs to be updated if any change from Data Flow
        self.VM_shapes = {
            "VM.Standard2.1",
            "VM.Standard2.2",
            "VM.Standard2.4",
            "VM.Standard2.8",
            "VM.Standard2.16",
            "VM.Standard2.24",  # not available in some tenancy
        }

    def __iter__(self):
        return self.list_apps().__iter__()

    def __len__(self):
        return len(self.list_apps())

    def _decorate_app(self, app):
        app.swagger_types["short_id"] = "str"
        app.ocid = app.id

        def to_dataframe(app_self):
            if "arguments" in application_attributes:
                application_attributes.remove("arguments")
            df = DataFrame.from_dict(
                {
                    key: getattr(app_self, key)
                    for key in application_attributes
                    if hasattr(app_self, key)
                },
                orient="index",
                columns=[""],
            )
            return df

        @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
        def show_in_notebook(app_self):
            """
            Describe the project by showing its properties
            """
            from IPython.display import display

            display(app_self)

        def _repr_html_(app_self):
            return (
                app_self.to_dataframe()
                .style.set_properties(**{"margin-left": "0px"})
                .render()
            )

        app.to_dataframe = MethodType(to_dataframe, app)
        app.show_in_notebook = MethodType(show_in_notebook, app)
        app._repr_html_ = MethodType(_repr_html_, app)

        return app

    def prepare_app(
        self,
        display_name: str,
        script_bucket: str,
        pyspark_file_path: str,
        spark_version: str = SPARK_VERSION.v2_4_4,
        compartment_id: str = None,
        archive_path: str = None,
        archive_bucket: str = None,
        logs_bucket: str = "dataflow-logs",
        driver_shape: str = "VM.Standard2.4",
        executor_shape: str = "VM.Standard2.4",
        num_executors: int = 1,
        arguments: list = [],
        script_parameters: dict = [],
    ) -> dict:
        """
        Check if the parameters provided by users to create an application are
        valid and then prepare app_configuration for creating an app or saving
        for future reuse.

        Parameters
        ----------
        display_name: str, required
            A user-friendly name. This name is not necessarily unique.
        script_bucket: str, required
            bucket in object storage to upload the pyspark file
        pyspark_file_path: str, required
            path to the pyspark file
        spark_version: str
            Allowed values are "2.4.4", "3.0.2".
        compartment_id: str
            OCID of the compartment to create a dataflow app. If not
            provided, compartment_id will use the same as the notebook session.
        archive_path: str, optional
            path to the archive file
        archive_bucket: str, optional
            bucket in object storage to upload the archive file
        logs_bucket: str, default is 'dataflow-logs'
            bucket in object storage to put run logs
        driver_shape: str
            The value to assign to the driver_shape property of this
            CreateApplicationDetails.
            Allowed values for this property are: "VM.Standard2.1",
            "VM.Standard2.2", "VM.Standard2.4", "VM.Standard2.8",
            "VM.Standard2.16", "VM.Standard2.24".
        executor_shape: str
            The value to assign to the executor_shape property of this
            CreateApplicationDetails.
            Allowed values for this property are: "VM.Standard2.1",
            "VM.Standard2.2", "VM.Standard2.4", "VM.Standard2.8",
            "VM.Standard2.16", "VM.Standard2.24".
        num_executors: int
            The number of executor VMs requested.
        arguments: list of str
            The values passed into the command line string to run the application
        script_parameters: dict
            The value of the parameters passed to the running application as
            command line arguments for the pyspark script.

        Returns
        -------
        app_configuration: dictionary containing all the validated params for CreateApplicationDetails.
        """
        if not self._check_bucket_exist(script_bucket):
            raise ValueError(
                "The bucket {} does not exist in object storage".format(script_bucket)
            )
        else:
            self.script_bucket = script_bucket

        if not self._check_bucket_exist(logs_bucket):
            raise ValueError(
                "The log bucket {} does not exist in object storage".format(logs_bucket)
            )
        else:
            self.logs_bucket_uri = (
                self.warehouse_bucket_uri
            ) = f"oci://{logs_bucket}@{self.namespace}"

        # check if local path of script file is valid
        self._check_valid_path(pyspark_file_path)

        if archive_path:
            # check if local path of archive file is valid
            self._check_valid_path(archive_path)

            if archive_bucket is None:
                # use script bucket by default if archive_bucket not provided
                archive_bucket = script_bucket
            else:
                if not self._check_bucket_exist(archive_bucket):
                    raise ValueError(
                        "The bucket {} does not exist in object storage".format(
                            archive_bucket
                        )
                    )

        # check whether the params have valid input type and value
        self._check_valid_param(
            display_name, driver_shape, executor_shape, num_executors
        )

        # when user try to specify a non-python application, we throw warnings
        if self.language is not None and self.language != "PYTHON":
            logger.warning("ADS only supports Python.")

        app_compartment_id = (
            self.compartment_id if compartment_id is None else compartment_id
        )

        app_configuration = {
            "compartment_id": app_compartment_id,
            "language": "PYTHON",
            "pyspark_file_path": pyspark_file_path,
            "script_bucket": self.script_bucket,
            "archive_path": archive_path,
            "archive_bucket": archive_bucket,
            "logs_bucket": logs_bucket,
            "display_name": self.display_name,
            "driver_shape": self.driver_shape,
            "executor_shape": self.executor_shape,
            "num_executors": self.num_executors,
            "spark_version": spark_version,
        }

        # here we handle the case where users specify arguments
        if arguments:
            # check if the arguments are valid
            for arg in arguments:
                if not isinstance(arg, str):
                    raise TypeError("Arguments must be a list of str.")

                if re.match("\$\{([^}]+)\}", arg):
                    arg_name = arg.strip("${}")
                    if " " in arg_name:
                        raise ValueError(
                            f"With {arg} in the format of "
                            "${var}, space is not allowed in "
                            f"{arg_name}"
                        )

                    if arg_name not in script_parameters:
                        logger.warning(
                            f"With `{arg}` in the format of "
                            "`${var}`, "
                            f"the argument `{arg_name}` will be replaced by the value provided in script parameters when passed in. "
                            f"While arguments not in this format are passed to the PySpark script verbatim."
                            f"Therefore, `{arg_name}` must be a valid key in script parameters."
                        )
                        raise KeyError(
                            f"{arg_name} doesn't exist in script parameters, thus {arg} is not valid."
                        )

            # convert script parameters to be a list of tuples
            app_configuration["script_parameters"] = [
                (k, script_parameters[k]) for k in script_parameters
            ]
            app_configuration["arguments"] = arguments

        return app_configuration

    def template(
        self,
        job_type: str = "standard_pyspark",
        script_str: str = "",
        file_dir: str = None,
        file_name: str = None,
    ) -> str:
        """
        Populate a prewritten pyspark or sparksql python script with
        user's choice to write additional lines and save in local directory.

        Parameters
        ----------
        job_type: str, default is 'standard_pyspark'
            Currently supports two types, 'standard_pyspark' or 'sparksql'
        script_str: str, optional, default is ''
            code provided by user to write in the python script
        file_dir: str, optional
            Directory to save the python script in local directory
        file_name: str, optional
            name of the python script to save to the local directory

        Returns
        -------
        script_path: str
            Path to the template generated python file in local directory
        """
        if file_dir is None:
            file_dir = self.dataflow_base_folder
            if not os.path.isdir(file_dir):
                os.mkdir(file_dir)

        if file_name is None:
            creation_time = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            file_name = f"{job_type}_{creation_time}.py"

        script_path = os.path.join(file_dir, file_name)

        if os.path.exists(script_path):
            logger.info(f"Overwriting {script_path}.")

        if job_type == "standard_pyspark":
            self._get_pyspark_template(script_path, script_str)
        elif job_type == "sparksql":
            self._get_sparksql_template(script_path, script_str)
        else:
            raise ValueError(
                "Currently only supports template for two job types, 'standard_pyspark' or 'sparksql'"
            )
        relative_path = os.path.relpath(script_path)
        # FileLink has to be tested with router to check compatibility. Till then let us comment it (ODSC-8310)
        # return display(FileLink(relative_path))
        logger.info(f"Code generated: {script_path}.")
        return script_path

    def _get_pyspark_template(self, script_path, script_str: str = ""):
        """
        Create a prewriiten pyspark script

        Parameters
        ----------
        script_path: str
            Path to the template generated python file in local directory
        script_str: str, optional, default is ''
            code provided by user to write in the python script

        Returns
        -------
        None
        """
        pyspark_template = env.get_template("dataflow_pyspark.jinja2")

        with open(script_path, "w") as fp:
            fp.write(pyspark_template.render(script_str=script_str))

    def _get_sparksql_template(self, script_path, script_str):
        """
        Create a prewriiten sparksql python script

        Parameters
        ----------
        script_path: str
            Path to the template generated python file in local directory
        script_str: str, optional, default is ''
            code provided by user to write in the python script

        Returns
        -------
        None
        """
        pyspark_template = env.get_template("dataflow_sparksql.jinja2")

        with open(script_path, "w") as fp:
            fp.write(pyspark_template.render(script_str=script_str))

    def _check_valid_path(self, file_path):
        """
        Returns
        -------
        valid_path: bool
            whether the provided file path is valid
        """
        file_dir = os.path.dirname(file_path)
        if len(file_dir) > 0 and not os.path.exists(file_dir):
            raise ValueError("The directoy of file {} does not exist".format(file_path))
        elif not os.path.exists(file_path):
            raise ValueError(
                " The directoy of the file {} is valid but the file does not exist".format(
                    file_path
                )
            )
        return True

    def _check_valid_param(
        self, display_name, driver_shape, executor_shape, num_executors
    ):
        """
        Returns
        -------
        valid_param: bool
            whether the params have valid input type and value
        """
        if not isinstance(display_name, str):
            raise TypeError("param 'display_name' must be string")
        else:
            self.display_name = display_name

        if not isinstance(driver_shape, str):
            raise TypeError("param 'driver_shape' must be string")
        elif driver_shape not in self.VM_shapes:
            raise ValueError("param 'driver_shape' is not a valid VM shape")
        else:
            self.driver_shape = driver_shape

        if not isinstance(executor_shape, str):
            raise TypeError("param 'executor_shape' must be string")
        elif executor_shape not in self.VM_shapes:
            raise ValueError("param 'executor_shape' is not a valid VM shape")
        else:
            self.executor_shape = executor_shape

        if not isinstance(num_executors, int):
            raise TypeError("param 'num_executors' must be an integer")
        elif num_executors < 1:
            raise ValueError("param 'num_executors' has a minimum value of 1")
        else:
            self.num_executors = num_executors
        return True

    def _check_bucket_exist(self, bucket_name: str) -> bool:
        """
        Returns
        -------
        bucket_exist: bool
            whether the bucket already exists in the object storage
        """
        try:
            bucket_response = self.object_storage_client.head_bucket(
                self.namespace, bucket_name
            )
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(
                    f"The bucket {bucket_name} does not exist in object storage"
                ) from se
            else:
                raise
        return True

    def _download(self, bucket_name, script_uri, target_folder):

        remote_pyspark_file_name = urlparse(script_uri).path[1:]

        local_pyspark_file_name = remote_pyspark_file_name
        if "/" in remote_pyspark_file_name:
            local_pyspark_file_name = remote_pyspark_file_name.replace("/", "_")
            if local_pyspark_file_name.endswith(".py"):
                local_pyspark_file_name = (
                    local_pyspark_file_name.rstrip(".py")
                    + "_"
                    + str(uuid.uuid4())[-6:]
                    + ".py"
                )
            else:
                local_pyspark_file_name += "_" + str(uuid.uuid4())[-6:]

        destination_pyspark_file = f"{target_folder}/{local_pyspark_file_name}"

        if not os.path.exists(destination_pyspark_file):

            with open(destination_pyspark_file, "wb") as f:

                f.write(
                    self.object_storage_client.get_object(
                        self.namespace, bucket_name, remote_pyspark_file_name
                    ).data.content
                )

            return destination_pyspark_file

        else:
            raise ValueError(
                f"The app script file ({remote_pyspark_file_name}) already exists in {target_folder}"
            )

    def _create_or_load_app(
        self,
        app_config: dict,
        file_uri: str,
        archive_uri: str = None,
        app_dir: str = None,
    ) -> object:

        with utils.get_progress_bar(2) as progress:
            progress.update()

            #
            # common to load & create
            #

            self.display_name = app_config["display_name"]
            self.driver_shape = app_config["driver_shape"]
            self.executor_shape = app_config["executor_shape"]
            self.num_executors = app_config["num_executors"]
            self.logs_bucket_uri = (
                self.warehouse_bucket_uri
            ) = f"oci://{app_config['logs_bucket']}@{self.namespace}"

            app_details = CreateApplicationDetails(
                compartment_id=app_config["compartment_id"],
                language="PYTHON",
                display_name=self.display_name,
                driver_shape=self.driver_shape,
                executor_shape=self.executor_shape,
                file_uri=file_uri,
                archive_uri=archive_uri,
                logs_bucket_uri=self.logs_bucket_uri,
                num_executors=self.num_executors,
                spark_version=app_config.get("spark_version", SPARK_VERSION.v2_4_4),
                warehouse_bucket_uri=self.warehouse_bucket_uri,
                arguments=app_config.get("arguments", []),
                parameters=[
                    ApplicationParameter(name=k, value=str(v))
                    for (k, v) in app_config.get("script_parameters", [])
                ],
            )

            new_app = self.df_client.create_application(app_details)
            progress.update("Done")
            # # add app_config to app obj as an attribute
            # new_app.data.configuration = app_config

            # make app dir
            if app_dir is None:
                app_dir = f"{self.dataflow_base_folder}/{self.display_name}_{str(uuid.uuid4())[-6:]}"

            pathlib.Path(app_dir).mkdir(parents=True, exist_ok=True)

            return DataFlowApp(
                app_config,
                new_app,
                app_dir,
                f"https://console."
                f"{self.region}.oraclecloud.com/data-flow/apps/details/{new_app.data.id}",
                os_auth=self.os_auth,
                df_auth=self.df_auth,
            )

    def create_app(
        self, app_config: dict, overwrite_script=False, overwrite_archive=False
    ) -> object:
        """
        Create a new dataflow application with the supplied app config.
        app_config contains parameters needed to create a new application,
        according to oci.data_flow.models.CreateApplicationDetails.

        Parameters
        ----------
        app_config: dict
            the config file that contains all necessary parameters used to create a dataflow app
        overwrite_script: bool
            whether to overwrite the existing pyscript script on Object Storage
        overwrite_archive: bool
            whether to overwrite the existing archive file on Object Storage

        Returns
        -------
        df_app: oci.dataflow.models.Application
            New dataflow application.
        """
        #
        # upload pyspark_file_path to OCI object storage
        #
        try:
            self._upload(
                app_config["pyspark_file_path"],
                app_config["script_bucket"],
                overwrite=overwrite_script,
            )
        except FileOverwriteError:
            raise ValueError(
                "You have a file with the same key in your bucket on object storage. Rename your file or set overwrite_script option to True."
            )
        script_name = os.path.basename(app_config["pyspark_file_path"])
        file_uri = f'oci://{app_config["script_bucket"]}@{self.namespace}/{script_name}'

        # upload archive file to object storage if specified
        if app_config["archive_path"] is None:
            return self._create_or_load_app(app_config, file_uri)
        else:
            try:
                self._upload(
                    app_config["archive_path"],
                    app_config["archive_bucket"],
                    overwrite=overwrite_archive,
                )
            except FileOverwriteError:
                raise ValueError(
                    "You have a file with the same key in your bucket on object storage. Rename your file or set overwrite_archive option to True."
                )
            archive_name = os.path.basename(app_config["archive_path"])
            archive_uri = (
                f'oci://{app_config["archive_bucket"]}@{self.namespace}/{archive_name}'
            )
            return self._create_or_load_app(
                app_config, file_uri, archive_uri=archive_uri
            )

    def _upload(self, local_path, bucket_name, overwrite=False):
        """
        upload local files to object storage

        Parameters
        ----------
        local_path: str
            the file path
        bucket_name: str
            bucket name on object storage to upload the file
        overwrite: bool
            whether to overwrite the existing file on Object Storage

        Returns
        -------
        None
        """
        object_name = os.path.basename(local_path)
        if self._check_object_exist(object_name, bucket_name):
            if not overwrite:
                raise FileOverwriteError()
            else:
                logger.warning(
                    "You have a file with the same key in your bucket on object storage. It will be overwritten per your request."
                )

        with open(local_path, "rb") as in_file:
            self.object_storage_client.put_object(
                self.namespace, bucket_name, object_name, in_file
            )
            logger.info(f"Finished uploading `{object_name}`.")

    def _check_object_exist(self, object_name: str, bucket_name: str) -> bool:
        """

        Parameters
        ----------
        object_name: str
            the file name on object storage
        bucket_name: str
            bucket name on object storage

        Returns
        -------
        bool
            whether the file already exists in the bucket in object storage
        """
        object_exist = self._check_object_exist_helper(
            object_name, bucket_name, start=None
        )

        if object_exist:
            logger.info(
                f"The file object `{object_name}` "
                f"already exists in bucket `{bucket_name}`."
            )
        else:
            logger.info(
                f"The file object `{object_name}` "
                f"does not exist in bucket `{bucket_name}` and will be uploaded."
            )
        return object_exist

    def _check_object_exist_helper(
        self, object_name: str, bucket_name: str, start: str = None
    ) -> bool:
        """

        Parameters
        ----------
        object_name: str
            the file name on object storage
        bucket_name: str
            bucket name on object storage
        start: str
            Object names returned by a list query must be greater or equal to this parameter.

        Returns
        -------
        bool
            whether the file already exists in the bucket in object storage
        """
        object_exist = False
        list_objects_response = self.object_storage_client.list_objects(
            self.namespace, bucket_name, start=start
        )

        objects_list = list_objects_response.data.objects
        for object_item in objects_list:
            if object_item.name == object_name:
                object_exist = True

        next_start_with = list_objects_response.data.next_start_with

        if object_exist or not next_start_with:
            return object_exist
        else:
            return self._check_object_exist_helper(
                object_name, bucket_name, start=next_start_with
            )

    def list_apps(
        self,
        include_deleted: bool = False,
        compartment_id: str = None,
        datetime_format: str = utils.date_format,
        **kwargs,
    ) -> object:
        """
        List all apps in a given compartment, or in the current notebook session's compartment.

        Parameters
        ----------
        include_deleted: bool, optional, default=False
            Whether to include deleted apps in the returned list.
        compartment_id: str, optional, default: NB_SESSION_COMPARTMENT_OCID
            The compartment specified to list apps.
        datetime_format: str, optional, default: '%Y-%m-%d %H:%M:%S'
            Change format for date time fields.

        Returns
        -------
        dsl: List
            List of Dataflow applications.
        """
        app_compartment_id = (
            self.compartment_id if compartment_id is None else compartment_id
        )
        list_applications_response = self.df_client.list_applications(
            app_compartment_id, **kwargs
        ).data

        # handle empty list
        if list_applications_response is None:
            logger.warning("No applications found.")
            return

        application_list_filtered = [
            self._decorate_app(app)
            for app in list_applications_response
            if include_deleted
            or Application.lifecycle_state != ApplicationSummary.LIFECYCLE_STATE_DELETED
        ]

        dsl = SummaryList(
            entity_list=application_list_filtered,
            datetime_format=datetime_format,
        )
        self.short_id_index.update(dsl.short_id_index)
        return dsl

    def get_app(self, app_id: str):
        """
        Get the Project based on app_id.

        Parameters
        ----------
        app_id: str, required
            The OCID of the dataflow app to get.

        Returns
        -------
        app: oci.dataflow.models.Application
            The oci.dataflow.models.Application with the matching ID.
        """
        if not app_id.startswith("ocid"):
            app_id = self.short_id_index[app_id]

        try:
            get_app_response = self.df_client.get_application(app_id)
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_app(get_app_response.data)

    def load_app(
        self,
        app_id: str,
        target_folder: str = None,
    ) -> object:
        """
        Load an existing dataflow application based on application id.
        The existing dataflow application can be created either from dataflow
        service or the dataflow integration of ADS.

        Parameters
        ----------
        app_id: str, required
            The OCID of the dataflow app to load.

        target_folder: str, optional,
            the folder to store the local artifacts of this application.
            If not specified, the target_folder will use the
            dataflow_base_folder by default.

        Returns
        -------
        dfa: ads.dataflow.dataflow.DataFlowApp
            A dataflow application of type ads.dataflow.dataflow.DataFlowApp
        """

        # support short id when loading an application by getting ocid based on
        # provided short id
        if not app_id.startswith("ocid"):
            app_id = self.short_id_index[app_id]

        # get app response that fetched using df client
        try:
            get_app_response = self.df_client.get_application(app_id)
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise

        # if users try to load a non-python application, we throw a warning
        if get_app_response.data.language != "PYTHON":
            logger.warning("ADS only supports Python.")
            return

        # for apps created with default logs bucket, logs_bucket_uri attribute in app response will be empty string
        # set default value manually for logs_bucket
        if get_app_response.data.logs_bucket_uri == "":
            logger.info(
                "Using the default logs bucket 'dataflow-logs'. Set the parameter `logs_bucket_uri` to use a different bucket."
            )
            logs_bucket = "dataflow-logs"
        else:
            logs_bucket = re.split(r"[@/]", get_app_response.data.logs_bucket_uri)[2]

        # reform app config from app response
        app_config = {
            "compartment_id": get_app_response.data.compartment_id,
            "language": get_app_response.data.language,
            "script_bucket": re.split(r"[@/]", get_app_response.data.file_uri)[2],
            "logs_bucket": logs_bucket,
            "archive_path": None,
            "archive_bucket": None,
            "display_name": get_app_response.data.display_name,
            "driver_shape": get_app_response.data.driver_shape,
            "executor_shape": get_app_response.data.executor_shape,
            "num_executors": get_app_response.data.num_executors,
            "spark_version": get_app_response.data.spark_version,
            "arguments": get_app_response.data.arguments,
            "script_parameters": [
                (param.name, param.value) for param in get_app_response.data.parameters
            ],
        }

        # set the default value to target_folder to dataflow_base_folder
        if target_folder is None:
            target_folder = self.dataflow_base_folder

        app_dir = (
            f"{target_folder}/{app_config['display_name']}_{str(uuid.uuid4())[-6:]}"
        )
        pathlib.Path(app_dir).mkdir(parents=True, exist_ok=True)

        app_config["pyspark_file_path"] = self._download(
            app_config["script_bucket"], get_app_response.data.file_uri, app_dir
        )

        if get_app_response.data.archive_uri != "":
            app_config["archive_bucket"] = re.split(
                r"[@/]", get_app_response.data.archive_uri
            )[2]
            app_config["archive_path"] = self._download(
                app_config["archive_bucket"], get_app_response.data.archive_uri, app_dir
            )

        return DataFlowApp(
            app_config,
            get_app_response,
            app_dir,
            f"https://console.{self.region}.oraclecloud.com/data-flow/apps/details/{get_app_response.data.id}",
            os_auth=self.os_auth,
            df_auth=self.df_auth,
        )


class DataFlowApp(DataFlow):
    @deprecated("2.6.3")
    def __init__(self, app_config, app_response, app_dir, oci_link, **kwargs):
        super().__init__(compartment_id=app_config["compartment_id"], **kwargs)
        self._config = app_config
        self.app_response = app_response
        self.app_dir = app_dir
        self._oci_link = oci_link

    def __iter__(self):
        return self.list_runs().__iter__()

    def __len__(self):
        return len(self.list_runs())

    def __repr__(self):
        return self._config["display_name"]

    def _decorate_run(self, run):
        run.swagger_types["short_id"] = "str"
        run.ocid = run.id

        def to_dataframe(run_self):
            if "arguments" in run_attributes:
                run_attributes.remove("arguments")
            df = DataFrame.from_dict(
                {
                    key: getattr(run_self, key)
                    for key in run_attributes
                    if hasattr(run_self, key)
                },
                orient="index",
                columns=[""],
            )
            return df

        @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
        def show_in_notebook(run_self):
            """
            Describe the project by showing it's properties
            """
            from IPython.display import display

            display(run_self)

        def _repr_html_(run_self):
            return (
                run_self.to_dataframe()
                .style.set_properties(**{"margin-left": "0px"})
                .render()
            )

        run.to_dataframe = MethodType(to_dataframe, run)
        run.show_in_notebook = MethodType(show_in_notebook, run)
        run._repr_html_ = MethodType(_repr_html_, run)

        return run

    @property
    def config(self) -> dict:
        """
        Retrieve the app_config file used to create the data flow app

        Returns
        -------
        app_config: Dict
            dictionary containing all the validated params for this DataFlowApp
        """
        return self._config

    @property
    def oci_link(self) -> object:
        """
        Retrieve the oci link of the data flow app

        Returns
        -------
        oci_link: str
            a link to the app page in an oci console.
        """
        return self._oci_link

    def prepare_run(
        self,
        run_display_name: str,
        compartment_id: str = None,
        logs_bucket: str = "",
        driver_shape: str = "VM.Standard2.4",
        executor_shape: str = "VM.Standard2.4",
        num_executors: int = 1,
        **kwargs,
    ) -> dict:
        """
        Check if the parameters provided by users to create a run are
        valid and then prepare run_config for creating run details.

        Parameters
        ----------
        run_display_name: str
            A user-friendly name. This name is not necessarily unique.
        compartment_id: str
            OCID of the compartment to create a dataflow run. If not
            provided, compartment_id will use the same as the dataflow app.
        logs_bucket: str
            bucket in object storage to put run logs, if not provided,
            will use the same logs_bucket as defined in app_config
        driver_shape: str
            The value to assign to the driver_shape property of this
            CreateApplicationDetails.
            Allowed values for this property are: "VM.Standard2.1",
            "VM.Standard2.2", "VM.Standard2.4", "VM.Standard2.8",
            "VM.Standard2.16", "VM.Standard2.24".
        executor_shape: str
            The value to assign to the executor_shape property of this
            CreateApplicationDetails.
            Allowed values for this property are: "VM.Standard2.1",
            "VM.Standard2.2", "VM.Standard2.4", "VM.Standard2.8",
            "VM.Standard2.16", "VM.Standard2.24".
        num_executors: int
            The number of executor VMs requested.

        Returns
        -------
        run_config: Dict
            Dictionary containing all the validated params for CreateRunDetails.
        """
        # if logs_bucket not provided to prepare run_config, will use the
        # same logs_bucket as defined in app_config
        if logs_bucket == "":
            logs_bucket = self.config["logs_bucket"]
        if not self._check_bucket_exist(logs_bucket):
            raise ValueError(
                "The log bucket {} does not exist in object storage".format(logs_bucket)
            )
        else:
            logs_bucket_uri = f"oci://{logs_bucket}@{self.namespace}"

        if not isinstance(run_display_name, str):
            raise TypeError("param 'run_display_name' must be string")

        if not isinstance(driver_shape, str):
            raise TypeError("param 'driver_shape' must be string")
        elif driver_shape not in self.VM_shapes:
            raise ValueError("param 'driver_shape' is not a valid VM shape")

        if not isinstance(executor_shape, str):
            raise TypeError("param 'executor_shape' must be string")
        elif executor_shape not in self.VM_shapes:
            raise ValueError("param 'executor_shape' is not a valid VM shape")

        if not isinstance(num_executors, int):
            raise TypeError("param 'num_executors' must be integer")
        elif num_executors < 1:
            raise ValueError("param 'num_executors' has a minimum value of 1")

        run_compartment_id = (
            self.compartment_id if compartment_id is None else compartment_id
        )

        run_configuration = {
            "compartment_id": run_compartment_id,
            "script_bucket": self._config["script_bucket"],
            "pyspark_file_path": self._config["pyspark_file_path"],
            "archive_path": self._config["archive_path"],
            "archive_bucket": self._config["archive_bucket"],
            "run_display_name": run_display_name,
            "logs_bucket": logs_bucket,
            "logs_bucket_uri": logs_bucket_uri,
            "driver_shape": driver_shape,
            "executor_shape": executor_shape,
            "num_executors": num_executors,
        }

        # get arguments from app config
        if "arguments" in self._config:
            run_configuration["arguments"] = list(self._config["arguments"])

        # get script parameters in app config
        if "script_parameters" in self._config:
            run_configuration["script_parameters"] = list(
                self._config["script_parameters"]
            )

        # update the new value of the argument in the run config when user provides new name value pairs
        if kwargs:
            # convert script parameters into dict
            param_dict = dict(run_configuration["script_parameters"])
            for param, value in kwargs.items():
                # overwrite the value
                if param in param_dict:
                    param_dict[param] = value
                else:
                    raise KeyError(f"{param} is not a valid key in script parameters")

            # convert param_dict back to list of tuples
            run_configuration["script_parameters"] = [
                (k, param_dict[k]) for k in param_dict
            ]

        return run_configuration

    def run(
        self,
        run_config: dict,
        save_log_to_local: bool = False,
        copy_script_to_object_storage: bool = True,
        copy_archive_to_object_storage: bool = True,
        pyspark_file_path: str = None,
        archive_path: str = None,
        wait: bool = True,
    ) -> object:
        """
        Create a new dataflow run with the supplied run config.
        run_config contains parameters needed to create a new run, according to oci.data_flow.models.CreateRunDetails.

        Parameters
        ----------
        run_config: dict, required
            The config file that contains all necessary parameters used to create a dataflow run
        save_log_to_local: bool, optional
            A boolean value that defaults to false. If set to true, it saves the log files to local dir
        copy_script_to_object_storage: bool, optional
            A boolean value that defaults to true. Local script will be copied to object storage
        copy_archive_to_object_storage: bool, optional
            A boolean value that defaults to true. Local archive file will be copied to object storage
        pyspark_file_path: str, optional
            The pyspark file path used for creating the dataflow app.
            if pyspark_file_path isn't specified then reuse the path that the app was created with.
        archive_path: str, optional
            The archive file path used for creating the dataflow app.
            if archive_path isn't specified then reuse the path that the app was created with.
        wait: bool, optional
            A boolean value that defaults to true.
            When True, the return will be ads.dataflow.dataflow.DataFlowRun in terminal state.
            When False, the return will be a ads.dataflow.dataflow.RunObserver.

        Returns
        -------
        df_run: Variable
            Either a new Data Flow run or a run observer.
        """

        if copy_script_to_object_storage:
            self._sync(pyspark_file_path)

        if run_config["archive_path"] is not None:
            if copy_archive_to_object_storage:
                self._sync(archive_path, type="archive")

        run_observer = RunObserver(self, run_config, save_log_to_local)

        if wait:
            return run_observer.wait()  # blocks, returns DataFlowRun
        else:
            return run_observer  # unblocks, returns RunObserver

    def _sync(self, file_path: str = None, type: str = "script") -> object:
        """
        Push to create a new app if the script has been modified.

        Parameters
        ----------
        file_path: str
            The pyspark file path used for creating the dataflow app.
            if pyspark_file_path isn't specified then reuse the path that the app was created with.
        type: str, only two types supported here, 'script' or 'archive'

        Returns
        -------
        Self, a Data Flow app object.
        """

        # local pyspark file, which may or may not be modified
        if type == "script":
            file_path = (
                self.config["pyspark_file_path"] if file_path is None else file_path
            )

            # compare byte content of two files
            # script no diff, return the original app obj
            if not self._modified(file_path, type="script"):
                return self

            os_bucket = self.config["script_bucket"]
            os_objectname = self.config["pyspark_file_path"].rsplit("/")[-1]

        elif type == "archive":
            file_path = self.config["archive_path"] if file_path is None else file_path
            if not self._modified(file_path, type="archive"):
                return self

            os_bucket = self.config["archive_bucket"]
            os_objectname = self.config["archive_path"].rsplit("/")[-1]

        # push file to object storage
        with open(file_path, "rb") as in_file:
            try:
                self.object_storage_client.put_object(
                    self.namespace,
                    os_bucket,
                    os_objectname,
                    in_file,
                )
                logger.info(
                    f"The existing file `{os_objectname}` in bucket "
                    f"`{os_bucket}` on object storage has been overwritten by your latest changes of "
                    f"`{os_objectname}`."
                )
            except ServiceError as se:
                if se.status == 404:
                    raise KeyError(se.message) from se
                else:
                    raise
        return self

    def _modified(
        self, file_path: str = None, type: str = "script", encoding="utf8"
    ) -> bool:
        """
        Check if any modification in the pyspark script

        Returns
        -------
        True or False
        """
        # read local python file into bytes
        if type == "script":
            bucket_name = self.config["script_bucket"]
            # remote pyspark file, which was pushed originally while creating this dataflow app
            remote_file_name = Path(self.config["pyspark_file_path"]).name.lstrip("/")
        elif type == "archive":
            bucket_name = self.config["archive_bucket"]
            remote_file_name = Path(self.config["archive_path"]).name.lstrip("/")
        local_file = open(file_path, "rb").read()

        # remote_file_name = Path(self.config["pyspark_file_path"]).name.lstrip('/')

        try:
            remote_file = self.object_storage_client.get_object(
                self.namespace, bucket_name, remote_file_name
            ).data.content
        except ServiceError as se:
            if se.code == "ObjectNotFound":
                local_filename = os.path.basename(file_path)
                logger.info(
                    f"The `{local_filename}` is "
                    f"not found in your bucket. "
                    f"The `{local_filename}` will be uploaded"
                )

                with open(file_path, "rb") as in_file:
                    self.object_storage_client.put_object(
                        self.namespace, bucket_name, local_filename, in_file
                    )

                return False
            else:
                raise se

        if type == "archive":
            return local_file != remote_file

        elif type == "script":
            if isinstance(local_file, bytes):
                local_file = local_file.decode(encoding)
            if isinstance(remote_file, bytes):
                remote_file = remote_file.decode(encoding)
            return local_file != remote_file

    def list_runs(
        self,
        include_failed: bool = False,
        datetime_format: str = utils.date_format,
        **kwargs,
    ) -> object:
        """
        List all run of a dataflow app

        Parameters
        ----------
        include_failed: bool, optional, default=False
            Whether to include failed runs in the returned list
        datetime_format: str, optional, default: '%Y-%m-%d %H:%M:%S'
            Change format for date time fields

        Returns
        -------
        df_runs: List
            List of Data flow runs.
        """

        list_runs_response = self.df_client.list_runs(
            self.compartment_id, **kwargs
        ).data

        # handle empty list
        if list_runs_response is None:
            logger.warning("No runs found.")
            return

        run_list_filtered = [
            self._decorate_run(run)
            for run in list_runs_response
            if include_failed
            or Run.lifecycle_state != RunSummary.LIFECYCLE_STATE_FAILED
        ]

        rsl = SummaryList(
            entity_list=run_list_filtered, datetime_format=datetime_format
        )
        self.short_id_index.update(rsl.short_id_index)
        return rsl

    def get_run(self, run_id: str):
        """
        Get the Run based on run_id

        Parameters
        ----------
        run_id: str, required
            The OCID of the dataflow run to get.

        Returns
        -------
        df_run: oci.dataflow.models.Run
            The oci.dataflow.models.Run with the matching ID.
        """
        if not run_id.startswith("ocid"):
            run_id = self.short_id_index[run_id]

        try:
            get_run_response = self.df_client.get_run(run_id)
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise
        return self._decorate_run(get_run_response.data)


class RunObserver:
    @deprecated("2.6.3")
    def __init__(self, app, run_config, save_log_to_local):
        self.app = app
        self._config = run_config
        self.save_log_to_local = save_log_to_local

        self._local_dir = self._create_run_dir()
        self._run_details = self._get_run_details()
        self._new_run = self.app.df_client.create_run(self._run_details)
        self._oci_link = f"https://console.{self.app.region}.oraclecloud.com/data-flow/runs/details/{self._new_run.data.id}"
        self._save_log_to_local = save_log_to_local

    def __repr__(self):
        return self._config["run_display_name"] + " " + self.status

    def wait(self):
        """
        Wait and monitor the run creation process.

        Parameters
        ----------
        None

        Returns
        -------
        df_run: oci.dataflow.models.Run
            The oci.dataflow.models.Run after monitoring is done.
        """
        # monitor the run creation process
        self._monitor_run()
        self._create_run_dir()

        return DataFlowRun(
            self._config,
            self._new_run,
            self.save_log_to_local,
            self._local_dir,
            os_auth=self.app.os_auth,
            df_auth=self.app.df_auth,
        )

    def _terminal_state(self, status):
        return status in ["SUCCEEDED", "FAILED", "CANCELED"]

    # this is a blocking function, it will only complete when dataflow run reaches terminal state
    def _monitor_run(self):
        curr_status = self.status
        if not self._terminal_state(curr_status):

            # when wait is called after the run being submitted, progress bar does not start from the beginning
            if curr_status == "ACCEPTED":
                progress_bar_num = 3
            elif curr_status == "IN_PROGRESS":
                progress_bar_num = 2
            elif curr_status == "SUCCEEDED":
                progress_bar_num = 1
            else:
                progress_bar_num = 4

            with utils.get_progress_bar(progress_bar_num) as progress:
                progress.update()
                while not self._terminal_state(curr_status):
                    time.sleep(2)
                    new_status = self.status
                    if new_status != curr_status:
                        progress.update(f"{new_status}")
                        curr_status = new_status
                progress.update("Done")

    def _create_run_dir(self):

        creation_time = datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        local_dir = (
            f"{self.app.app_dir}/{self._config['run_display_name']}-{creation_time}"
        )
        pathlib.Path(local_dir).mkdir(parents=True, exist_ok=True)

        # add pyspark script for the run in the run dir
        user_pyspark_file_path = self._config["pyspark_file_path"]
        filename = os.path.basename(user_pyspark_file_path)
        run_pyspark_file_path = os.path.join(local_dir, filename)

        # copy content
        from shutil import copyfile

        copyfile(user_pyspark_file_path, run_pyspark_file_path)

        # add run metadata for the run in the run dir
        run_config_path = os.path.join(local_dir, "run_metadata.json")
        with open(run_config_path, "w") as fp:
            json.dump(self._config, fp)

        return local_dir

    def _get_run_details(self):
        self.run_display_name = self._config["run_display_name"]
        self.logs_bucket_uri = self._config["logs_bucket_uri"]
        self.driver_shape = self._config["driver_shape"]
        self.executor_shape = self._config["executor_shape"]
        self.num_executors = self._config["num_executors"]

        run_details = CreateRunDetails(
            application_id=self.app.app_response.data.id,
            compartment_id=self._config["compartment_id"],
            display_name=self.run_display_name,
            logs_bucket_uri=self.logs_bucket_uri,
            driver_shape=self.driver_shape,
            executor_shape=self.executor_shape,
            num_executors=self.num_executors,
            arguments=self._config.get("arguments", []),
            parameters=[
                ApplicationParameter(name=k, value=str(v))
                for (k, v) in self._config.get("script_parameters", [])
            ],
        )
        return run_details

    @property
    def status(self) -> str:
        """
        Returns the lifecycle state of the Data Flow run
        """
        return self.app.df_client.get_run(self._new_run.data.id).data.lifecycle_state

    @property
    def config(self) -> dict:
        """
        Retrieve the run_config file used to create the data flow run

        Returns
        -------
        run_config: Dict
            Dictionary containing all the validated parameters for this Data Flow run
        """
        return self._config

    def update_config(self, param_dict) -> None:
        """
        Modify the run_config file used to create the data flow run

        Parameters
        ----------
        param_dict: Dict
            dictionary containing the key value pairs of the run_config parameters and
            the updated values.

        Returns
        -------
        None
        """
        for key, val in param_dict.items():
            if key in self._config:
                self._config[key] = val
            else:
                raise ValueError(
                    "The key '{}' does not exist in run_config parameters".format(key)
                )

    @property
    def local_dir(self) -> str:
        """
        Retrieve the local directory of the data flow run

        Returns
        -------
        local_dir: str
            the local path to the Data Flow run
        """
        return self._local_dir

    @property
    def oci_link(self) -> object:
        """
        Retrieve the oci link of the data flow run

        Returns
        -------
        oci_link: str
            link to the run page in an oci console
        """
        return self._oci_link


class DataFlowRun(DataFlow):
    LOG_OUTPUTS = ["stdout", "stderr"]

    @deprecated("2.6.3")
    def __init__(
        self, run_config, run_response, save_log_to_local, local_dir, **kwargs
    ):
        super().__init__(compartment_id=run_config["compartment_id"], **kwargs)
        self._config = run_config
        self.run_response = run_response
        self.run_id = self.run_response.data.id
        self._log_stdout = None
        self._log_stderr = None
        self._local_dir = local_dir
        self._status = self.df_client.get_run(self.run_id).data.lifecycle_state
        self._oci_link = None
        if save_log_to_local:
            for log_type in DataFlowRun.LOG_OUTPUTS:
                self.fetch_log(log_type).save()

    def __repr__(self):
        return self._config["run_display_name"]

    @property
    def config(self) -> dict:
        """
        Retrieve the run_config file used to create the Data Flow run

        Returns
        -------
        run_config: Dict
            dictionary containing all the validated params for this DataFlowRun
        """
        return self._config

    def update_config(self, param_dict) -> None:
        """
        Modify the run_config file used to create the data flow run

        Parameters
        ----------
        param_dict: Dict
            Dictionary containing the key value pairs of the run_config parameters and
            the updated values.

        Returns
        -------
        None
        """
        for key, val in param_dict.items():
            if key in self._config:
                self._config[key] = val
            else:
                raise ValueError(
                    "The key '{}' does not exist in run_config parameters".format(key)
                )

    @property
    def status(self) -> str:
        """
        Retrieve the status of the data flow run

        Returns
        -------
        status: str
            String that describes the status of the run
        """
        return self._status

    @property
    def log_stdout(self) -> object:
        """
        Retrieve the stdout of the data flow run

        Returns
        -------
        log_out: ads.dataflow.dataflow.DataFlowLog
            a clickable link that opens the stdout log in another tab in a JupyterLab notebook environment
        """
        if self._log_stdout is None:
            self._log_stdout = self.fetch_log("stdout")
        return self._log_stdout

    @property
    def log_stderr(self) -> object:
        """
        Retrieve the stderr of the data flow run

        Returns
        -------
        log_error: ads.dataflow.dataflow.DataFlowLog
            a clickable link that opens the stderror log in another tab in jupyter notebook environment
        """
        if self._log_stderr is None:
            self._log_stderr = self.fetch_log("stderr")
        return self._log_stderr

    @property
    def local_dir(self) -> str:
        """
        Retrieve the local directory of the data flow run

        Returns
        -------
        local_dir: str
            the local path to the Data Flow run
        """
        return self._local_dir

    @property
    def oci_link(self) -> object:
        """
        Retrieve the oci link of the data flow run

        Returns
        -------
        oci_link: str
            link to the run page in an oci console
        """
        return self._oci_link

    def fetch_log(self, log_type: str) -> object:
        """
        Fetch the log information of a run

        Parameters
        ----------
        log_type: str, have two values, 'stdout' or 'stderr'

        Returns
        -------
        dfl: DataFlowLog
            a Data Flow log object
        """
        if log_type not in DataFlowRun.LOG_OUTPUTS:
            raise ValueError(
                f"Invalid log type ({log_type}), valid types: {', '.join(DataFlowRun.LOG_OUTPUTS)}"
            )

        tmp = self.df_client.get_run_log(
            self.run_id, f"spark_application_{log_type}.log.gz"
        )
        text_str = str(tmp.data.text.lstrip("\x00").rstrip("\n"))

        opc_request_id = self.run_response.data.opc_request_id

        log_filename = os.path.join(
            opc_request_id, f"spark_application_{log_type}.log.gz"
        )
        object_storage_log_path = os.path.join(
            self._config["logs_bucket_uri"], log_filename
        )
        log_local_dir = os.path.join(self._local_dir, "logs")

        log_obj = DataFlowLog(text_str, object_storage_log_path, log_local_dir)
        if log_type == "stdout":
            self._log_stdout = log_obj
        elif log_type == "stderr":
            self._log_stderr = log_obj

        return log_obj

    def _repr_html_(self):
        if self.status == "FAILED":
            logger.warning("Run failed. See the logs for details.")
            logger.info("Printing tail of the stdout log...")
            self.log_stdout.tail()
        else:
            return


class DataFlowLog:
    @deprecated("2.6.3")
    def __init__(self, text, oci_path, log_local_dir):
        self.text = str(text)
        self._oci_path = oci_path
        self._local_dir = log_local_dir
        self._local_path = None
        self.line_list = self.text.split("\n")

    def head(self, n: int = 10):
        """
        Show the first n lines of the log as the output of the notebook cell

        Parameters
        ----------
        n: int, default is 10
            the number of lines from head of the log file

        Returns
        -------
        None
        """
        for _, v in enumerate(self.line_list[:n]):
            print(v)

    def tail(self, n: int = 10):
        """
        Show the last n lines of the log as the output of the notebook cell

        Parameters
        ----------
        n: int, default is 10
            the number of lines from tail of the log file

        Returns
        -------
        None
        """
        for _, v in enumerate(self.line_list[-n:]):
            print(v)

    def show_all(self):
        """
        Show all content of the log as the output of the notebook cell

        Returns
        -------
        None
        """
        for _, v in enumerate(self.line_list):
            print(v)

    @property
    def oci_path(self):
        """
        Get the path of the log file in object storage

        Returns
        -------
        oci_path: str
            Path of the log file in object storage
        """
        return self._oci_path

    @property
    def local_path(self):
        """
        Get the path of the log file in local directory

        Returns
        -------
        local_path: str
            Path of the log file in local directory
        """
        if self._local_path is None:
            logger.warning(
                "The log file is not stored in local directory. "
                "Call the save() method to save the log file to local first.",
            )

        return self._local_path

    @property
    def local_dir(self):
        """
        Get the local directory where the log file is saved.

        Returns
        -------
        local_dir: str
            Path to the local directory where the log file is saved.
        """
        return self._local_dir

    def save(self, log_dir=None):
        """
        save the log file to a local directory.

        Parameters
        ----------
        log_dir: str,
            The path to the local directory to save log file, if not
        set, log will be saved to the _local_dir by default.

        Returns
        -------
        None
        """
        if self._local_path is not None:
            logger.warning(f"The log file is already exists in `{self._local_path}`.")
            return
        if log_dir is not None:
            self._local_dir = log_dir

        if not os.path.isdir(self._local_dir):
            os.makedirs(self._local_dir)

        filename = os.path.basename(self._oci_path)
        self._local_path = os.path.join(self._local_dir, filename)

        with open(self._local_path, "w+") as in_file:
            in_file.write(self.text)

        logger.info(f"The log file saved to `{self._local_path}`.")

    def __str__(self):
        return self.text

    def _repr_html_(self):
        """
        Display the link of the log file to open in another tab of the jupyter
        environment.

        Returns
        -------
        link of the log file
        """
        if self._local_path is None:
            return

        relative_path = os.path.relpath(self._local_path)
        # FileLink has to be tested with router to check compatibility. Till then let us comment it (ODSC-8310)
        # return display(FileLink(relative_path))
        return logger.info(f"Log file saved to {self._local_path}.")
